import contextlib
import functools
import logging
import uuid
from typing import Optional
import concurrent.futures
from typing import Any
from collections.abc import Iterable

from opentelemetry import trace

from .base import (
    BaseJobExecutor, AsyncAdapter,
)
from libertem.common.executor import (
    JobCancelledError, TaskProtocol, TaskCommHandler, WorkerContext,
    SimpleWorkerQueue, WorkerQueue, Environment
)
from libertem.common.async_utils import sync_to_async
from libertem.utils.devices import detect
from libertem.common.scheduler import Worker, WorkerSet
from libertem.common.backend import get_use_cuda
from libertem.common.tracing import TracedThreadPoolExecutor


log = logging.getLogger(__name__)
tracer = trace.get_tracer(__name__)


class ConcurrentWorkerContext(WorkerContext):
    def __init__(self, msg_queue: WorkerQueue):
        self._msg_queue = msg_queue

    def signal(self, ident: str, topic: str, msg_dict: dict[str, Any]):
        if 'ident' in msg_dict:
            raise ValueError('ident is a reserved name')
        msg_dict.update({'ident': ident})
        self._msg_queue.put((topic, msg_dict))


def _run_task(task, params, task_id, threaded_executor, msg_queue, scatter_map):
    """
    Wraps the task to be run in the pool
    """
    params = scatter_map[params]
    worker_context = ConcurrentWorkerContext(msg_queue)
    env = Environment(threads_per_worker=1,
                      threaded_executor=threaded_executor,
                      worker_context=worker_context)
    task_result = task(env=env, params=params)
    return {
        "task_result": task_result,
        "task_id": task_id,
    }


class ConcurrentJobExecutor(BaseJobExecutor):
    '''
    :class:`JobExecutor` that uses :mod:`python.concurrent.futures`.

    .. versionadded:: 0.9.0

    Parameters
    ----------

    client : concurrent.futures.Executor
    is_local : bool
        Shut the client down when the executor closes.
    '''
    def __init__(self, client: concurrent.futures.Executor, is_local=False):
        # Only import if actually instantiated, i.e. will likely be used
        import libertem.preload  # noqa: 401
        self.is_local = is_local
        self.client = client
        self._futures = {}
        self._scatter_map = {}

    @contextlib.contextmanager
    def scatter(self, obj):
        handle = str(uuid.uuid4())
        self._scatter_map[handle] = obj
        try:
            yield handle
        finally:
            del self._scatter_map[handle]

    def scatter_update(self, handle, obj):
        self._scatter_map[handle] = obj

    def scatter_update_patch(self, handle, patch):
        self._scatter_map[handle].patch(patch)

    def _get_future(self, wrapped_task, idx, params_handle, msg_queue):
        return self.client.submit(
            _run_task,
            task=wrapped_task, params=params_handle, task_id=idx,
            threaded_executor=True, msg_queue=msg_queue,
            scatter_map=self._scatter_map,
        )

    def run_tasks(
        self,
        tasks: Iterable[TaskProtocol],
        params_handle: Any,
        cancel_id: Any,
        task_comm_handler: TaskCommHandler,
    ):
        tasks = list(tasks)

        def _id_to_task(task_id):
            return tasks[task_id]

        self._futures[cancel_id] = []

        msg_queue = SimpleWorkerQueue()

        for idx, wrapped_task in list(enumerate(tasks)):
            future = self._get_future(wrapped_task, idx, params_handle, msg_queue)
            self._futures[cancel_id].append(future)

        with task_comm_handler.monitor(msg_queue):
            try:
                as_completed = concurrent.futures.as_completed(self._futures[cancel_id])
                for future in as_completed:
                    result_wrap = future.result()
                    if future.cancelled():
                        del self._futures[cancel_id]
                        raise JobCancelledError()
                    result = result_wrap['task_result']
                    task = _id_to_task(result_wrap['task_id'])
                    yield result, task
            finally:
                if cancel_id in self._futures:
                    del self._futures[cancel_id]

    def cancel(self, cancel_id):
        if cancel_id in self._futures:
            for future in self._futures[cancel_id]:
                future.cancel()

    def run_function(self, fn, *args, **kwargs):
        """
        run a callable :code:`fn` on any worker
        """
        fn_with_args = functools.partial(fn, *args, **kwargs)
        future = self.client.submit(fn_with_args)
        return future.result()

    def map(self, fn, iterable):
        """
        Run a callable :code:`fn` for each element in :code:`iterable`, on arbitrary worker nodes.

        Parameters
        ----------

        fn : callable
            Function to call. Should accept exactly one parameter.

        iterable : Iterable
            Which elements to call the function on.
        """
        return self.client.map(fn, iterable)

    def get_available_workers(self):
        resources = {"compute": 1, "CPU": 1}
        if get_use_cuda() is not None:
            resources["CUDA"] = 1
            return WorkerSet([
                Worker(name='concurrent', host='localhost', resources=resources, nthreads=1)
            ])
        else:
            devices = detect()
            return WorkerSet([
                Worker(
                    name='concurrent', host='localhost',
                    resources=resources,
                    nthreads=len(devices['cpus']),
                )
            ])

    def run_each_host(self, fn, *args, **kwargs):
        """
        Run a callable :code:`fn` once on each host, gathering all results into
        a dict host -> result
        """
        # TODO: any cancellation/errors to handle?
        future = self.client.submit(fn, *args, **kwargs)
        return {"localhost": future.result()}

    def run_each_worker(self, fn, *args, **kwargs):
        future = self.client.submit(fn, *args, **kwargs)
        return {"inline": future.result()}

    def close(self):
        if self.is_local:
            self.client.shutdown(wait=False)

    @classmethod
    def make_local(cls, n_threads: Optional[int] = None):
        """
        Create a local ConcurrentJobExecutor backed by
        a :class:`python:concurrent.futures.ThreadPoolExecutor`

        Parameters
        ----------

        n_threads : Optional[int]
            The number of threads to spawn in the executor,
            by default None in which case as many threads as there
            are CPU cores will be spawned.

        Returns
        -------
        ConcurrentJobExecutor
            the connected JobExecutor
        """
        if n_threads is None:
            devices = detect()
            n_threads = len(devices['cpus'])
        client = TracedThreadPoolExecutor(tracer, max_workers=n_threads)
        return cls(client=client, is_local=True)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class AsyncConcurrentJobExecutor(AsyncAdapter):
    def __init__(self, wrapped=None, *args, **kwargs):
        if wrapped is None:
            wrapped = ConcurrentJobExecutor(*args, **kwargs)
        super().__init__(wrapped)

    @classmethod
    async def make_local(cls):
        executor = await sync_to_async(functools.partial(
            ConcurrentJobExecutor.make_local,
        ))
        return cls(wrapped=executor)



import numpy as np
import pytest

import libertem.masks as m


def test_background_subtraction():
    mask = m.background_subtraction(10, 10, 20, 20, 5, 3)
    assert np.allclose(np.sum(mask), 0)


@pytest.mark.parametrize(
    'antialiased', (True, False)
)
def test_radial_gradient(antialiased):
    template = m.radial_gradient(
        centerX=30, centerY=33,
        imageSizeX=65,
        imageSizeY=66,
        radius=17,
        antialiased=antialiased
    )

    assert template.shape == (66, 65)

    for i in range(17):
        assert np.allclose(template[33, 30+i], i/17)
        assert np.allclose(template[33, 30-i], i/17)
        assert np.allclose(template[33+i, 30], i/17)
        assert np.allclose(template[33-i, 30], i/17)

    for i in range(18, 30):
        assert np.allclose(template[33, 30+i], 0)
        assert np.allclose(template[33, 30-i], 0)
        assert np.allclose(template[33+i, 30], 0)
        assert np.allclose(template[33-i, 30], 0)


def test_sparse_template_multi_stack():
    template = np.ones((2, 3))
    stack = m.sparse_template_multi_stack(
        mask_index=(0, 1, 2),
        offsetY=(13, 14, 15),
        offsetX=(15, 14, 13),
        template=template,
        imageSizeY=32,
        imageSizeX=32
    )
    t1 = np.zeros((32, 32))
    t1[13:15, 15:18] = 1

    t2 = np.zeros((32, 32))
    t2[14:16, 14:17] = 1

    t3 = np.zeros((32, 32))
    t3[15:17, 13:16] = 1

    assert np.allclose(stack[0].todense(), t1)
    assert np.allclose(stack[1].todense(), t2)
    assert np.allclose(stack[2].todense(), t3)


def test_radial_bins():
    bins = m.radial_bins(35, 37, 80, 80, n_bins=42)
    assert np.allclose(1, bins.sum(axis=0).todense())
    assert bins.shape == (42, 80, 80)


def test_radial_bins_dense():
    bins = m.radial_bins(40, 41, 80, 80, n_bins=2)
    assert np.allclose(1, bins.sum(axis=0))
    assert bins.shape == (2, 80, 80)


def test_oval_radial_background_balance():
    radius, angle = m.polar_map(
        centerX=10, centerY=10,
        imageSizeX=21, imageSizeY=21,
        stretchY=1.5,
        angle=np.pi/4
    )
    template = m.radial_gradient_background_subtraction(
        r=radius,
        r0=5,
        r_outer=7
    )
    template = m.balance(template)
    outside = radius > 7
    inside = (radius < 5) * (radius > 0)

    assert np.allclose(template.sum(), 0)
    assert np.allclose(template[outside], 0)
    assert np.all(template[inside] > 0)

    assert np.allclose(template[0, :], 0)
    assert np.allclose(template[20, :], 0)
    assert np.allclose(template[:, 0], 0)
    assert np.allclose(template[:, 20], 0)

    diag = int(np.sqrt(5))

    # Confirm stretched gradient in 45 diagonal
    assert template[10+diag, 10+diag] > 0
    assert template[10-diag, 10+diag] < template[10+diag, 10+diag]


def test_oval_radial_background_symmetry():
    radius, angle = m.polar_map(
        centerX=10, centerY=10,
        imageSizeX=21, imageSizeY=21,
        stretchY=1.5,
        angle=0.
    )

    radius2, angle2 = m.polar_map(
        centerX=10, centerY=10,
        imageSizeX=21, imageSizeY=21,
        stretchY=1.5,
        angle=np.pi
    )

    radius3, angle3 = m.polar_map(
        centerX=10, centerY=10,
        imageSizeX=21, imageSizeY=21,
        stretchY=1.5,
        angle=np.pi/4
    )

    radius4, angle4 = m.polar_map(
        centerX=10, centerY=10,
        imageSizeX=21, imageSizeY=21,
        stretchY=1.5,
        angle=-np.pi/4
    )

    assert np.allclose(radius, radius2)
    assert np.allclose(radius3, np.flip(radius4, axis=1))


def test_rectmask():
    rect = np.ones([3, 3])
    testrect = np.zeros([5, 5])
    rect1 = 1*testrect
    rect1[0:3, 0:3] = rect
    rect2 = 1*testrect
    rect2[0:3, 2:5] = rect
    rect3 = 1*testrect
    rect3[2:5, 0:3] = rect
    rect4 = 1*testrect
    rect4[2:5, 2:5] = rect
    assert np.allclose(m.rectangular(2, 2, 3, 3, 5, 5), rect4)
    assert np.allclose(m.rectangular(2, 2, -3, 3, 5, 5), rect3)
    assert np.allclose(m.rectangular(2, 2, 3, -3, 5, 5), rect2)
    assert np.allclose(m.rectangular(2, 2, -3, -3, 5, 5), rect1)


def test_empty_rectmask():
    assert not np.any(m.rectangular(2, 2, 0, 3, 5, 5))
    assert not np.any(m.rectangular(2, 2, 3, 0, 5, 5))
    assert not np.any(m.rectangular(2, 2, 0, 0, 5, 5))



from .base import BaseAnalysis, AnalysisResultSet, AnalysisResult
from libertem.udf.masks import ApplyMasksUDF
from libertem.analysis.getroi import get_roi


class BaseMasksAnalysis(BaseAnalysis):
    """
    Base class for any masks-based analysis; you only need to implement
    ``get_results``, ``get_udf_results`` and ``get_mask_factories``.
    Overwrite  ``get_use_sparse`` to return True to calculate with sparse mask matrices.

    .. versionchanged:: 0.4.0
        Add support to use this Analysis with both ApplyMasksJob and ApplyMasksUDF :issue:`549`

    .. versionchanged:: 0.7.0
        ApplyMasksJob support removed
    """
    def get_udf(self):
        return ApplyMasksUDF(
            mask_factories=self.get_mask_factories(),
            use_sparse=self.get_use_sparse(),
            mask_count=self.get_preset_mask_count(),
            mask_dtype=self.get_preset_mask_dtype(),
            preferred_dtype=self.get_preset_dtype()
        )

    def get_mask_factories(self):
        raise NotImplementedError()

    def get_use_sparse(self):
        return self.parameters.get('use_sparse', None)

    def get_preset_mask_count(self):
        return self.parameters.get('mask_count', None)

    def get_preset_mask_dtype(self):
        return self.parameters.get('mask_dtype', None)

    def get_preset_dtype(self):
        return self.parameters.get('dtype', None)


class SingleMaskAnalysis(BaseMasksAnalysis):
    def get_udf_results(self, udf_results, roi, damage):
        data = udf_results['intensity'].data
        return self.get_generic_results(data[..., 0], damage=damage)

    def get_description(self):
        raise NotImplementedError

    def get_generic_results(self, data, damage):
        from libertem.viz import visualize_simple
        if data.dtype.kind == 'c':
            return SingleMaskResultSet(
                self.get_complex_results(
                    data,
                    key_prefix='intensity',
                    title='intensity',
                    desc=self.get_description(),
                    damage=damage,
                )
            )
        return SingleMaskResultSet([
            AnalysisResult(
                raw_data=data,
                visualized=visualize_simple(data, damage=damage),
                key="intensity", title="intensity [lin]",
                desc=f'{self.get_description()} lin-scaled'
            ),
            AnalysisResult(
                raw_data=data,
                visualized=visualize_simple(data, logarithmic=True, damage=damage),
                key="intensity_log", title="intensity [log]",
                desc=f'{self.get_description()} log-scaled'
            ),
        ])


class MasksResultSet(AnalysisResultSet):
    """
    Running a :class:`MasksAnalysis` via :meth:`libertem.api.Context.run` on a dataset
    returns an instance of this class.

    If any of the masks or the dataset contain complex numbers, the regular mask results
    attributes carry the absolute value of the results, and additional attributes with real
    part, imaginary part, phase and full complex result are available.

    .. versionadded:: 0.3.0

    Attributes
    ----------
    mask_0, mask_1, ..., mask_<n> : libertem.analysis.base.AnalysisResult
        For dataset and masks containing only real numbers:
        Results of the element-wise multiplication and sum of each individual mask with
        each detector frame. Each mask result has the shape of the navigation dimension.
        These keys contain the absolute value of the result if dataset or masks contain
        complex numbers.
    mask_0, mask_0_real, mask_0_imag, mask_0_angle, mask_0_complex,\
    mask_1, mask_1_real, ...,\
    mask_<n>, ..., mask_<n>_complex : libertem.analysis.base.AnalysisResult
        If masks or dataset contain complex numbers: Absolute, real part, imaginary part,
        phase angle, complex result of the element-wise multiplication and sum of each individual
        mask with each detector frame. Each mask result has the shape of the navigation dimension.
    """
    pass


class SingleMaskResultSet(AnalysisResultSet):
    """
    A number of Analyses that are based on applying a single mask create an instance of this class
    as a result when executed via :meth:`libertem.api.Context.run`.

    If the dataset contains complex numbers, the regular result attribute carries the
    absolute value of the result, and additional attributes with real part, imaginary part,
    phase and full complex result are available.

    .. versionadded:: 0.3.0

    Attributes
    ----------
    intensity : libertem.analysis.base.AnalysisResult
        Sum of the selected region for each detector frame, with shape of
        the navigation dimension. Absolute of the result if the dataset or mask contains
        complex numbers.
    intensity_log : libertem.analysis.base.AnalysisResult
        Log-scaled sum of the selected region for each detector frame, with shape of
        the navigation dimension. Absolute of the result if the dataset or mask contains
        complex numbers.

        .. versionadded:: 0.6.0

    intensity_real : libertem.analysis.base.AnalysisResult
        Real part of the sum of the selected region. This is only available if the dataset
        or mask contains complex numbers.
    intensity_imag : libertem.analysis.base.AnalysisResult
        Imaginary part of the sum of the selected region. This is only available if the dataset
        contains complex numbers.
    intensity_angle : libertem.analysis.base.AnalysisResult
        Phase angle of the sum of the selected region. This is only available if the dataset
        or mask contains complex numbers.
    intensity_complex : libertem.analysis.base.AnalysisResult
        Complex result of the sum of the selected region. This is only available if the dataset
        or mask contains complex numbers.
    """
    pass


class MasksAnalysis(BaseMasksAnalysis):
    TYPE = 'UDF'

    def get_mask_factories(self):
        return self.parameters['factories']

    def get_generic_results(self, data, damage):
        from libertem.viz import visualize_simple
        if data.dtype.kind == 'c':
            results = []
            for idx in range(data.shape[-1]):
                results.extend(
                    self.get_complex_results(
                        data[..., idx],
                        key_prefix="mask_%d" % idx,
                        title="mask %d" % idx,
                        desc="integrated intensity for mask %d" % idx,
                        damage=damage
                    )
                )
            return MasksResultSet(results)
        return MasksResultSet([
            AnalysisResult(
                raw_data=data[..., idx],
                visualized=visualize_simple(data[..., idx], damage=damage),
                key="mask_%d" % idx,
                title="mask %d" % idx,
                desc="integrated intensity for mask %d" % idx)
            for idx in range(data.shape[-1])
        ])

    def get_roi(self):
        return get_roi(params=self.parameters, shape=self.dataset.shape.nav)

    def get_udf_results(self, udf_results, roi, damage):
        data = udf_results['intensity'].data
        return self.get_generic_results(data, damage=damage)



import json
import uuid
import urllib.parse as urlparse
from urllib.parse import urlencode

from utils import assert_msg


def _get_raw_params(path):
    return {
        "dataset": {
            "params": {
                "type": "raw",
                "path": path,
                "name": "name.raw",
                "dtype": "float32",
                "sig_shape": [128, 128],
                "enable_direct": False,
                "nav_shape": [16, 16],
                "sync_offset": 0
            }
        }
    }


def _get_uuid_str():
    return str(uuid.uuid4())


def add_token(url, token):
    if token is None:
        return url
    url_parts = list(urlparse.urlparse(url))
    query = dict(urlparse.parse_qsl(url_parts[4]))
    query['token'] = token
    url_parts[4] = urlencode(query)
    return urlparse.urlunparse(url_parts)


async def create_connection(base_url, http_client, scheduler_url=None, token=None):
    conn_url = f"{base_url}/api/config/connection/"
    conn_url = add_token(conn_url, token)
    if scheduler_url is None:
        conn_details = {
            'connection': {
                'type': 'local',
                'numWorkers': 2,
                'cudas': {},
            }
        }
    else:
        conn_details = {
            'connection': {
                'type': 'tcp',
                'address': scheduler_url,
            }
        }
    print("checkpoint 0")
    async with http_client.put(conn_url, json=conn_details) as response:
        assert response.status == 200
        response_json = await response.json()
        assert response_json['status'] == 'ok'


async def consume_task_results(ws, job_uuid):
    num_followup = 0
    done = False
    while not done:
        msg = json.loads(await ws.recv())
        if msg['messageType'] == 'TASK_RESULT':
            assert_msg(msg, 'TASK_RESULT')
            assert msg['job'] == job_uuid
        elif msg['messageType'] == 'FINISH_JOB':
            done = True  # but we still need to check followup messages below
        elif msg['messageType'] == 'JOB_PROGRESS':
            assert msg['job'] == job_uuid
        elif msg['messageType'] == 'JOB_ERROR':
            raise Exception('JOB_ERROR: {}'.format(msg['msg']))
        else:
            raise Exception("invalid message type: {}".format(msg['messageType']))

        if 'followup' in msg:
            for i in range(msg['followup']['numMessages']):
                msg = await ws.recv()
                # followups should be PNG encoded:
                assert msg[:8] == b'\x89\x50\x4E\x47\x0D\x0A\x1A\x0A'
                num_followup += 1
    assert num_followup > 0


async def create_default_dataset(default_raw, ws, http_client, base_url, token=None):
    ds_uuid = _get_uuid_str()
    ds_url = "{}/api/datasets/{}/".format(
        base_url, ds_uuid
    )
    ds_url = add_token(ds_url, token)
    ds_data = _get_raw_params(default_raw._path)

    print("creating dataset: dispatching PUT request")

    async with http_client.put(ds_url, json=ds_data) as resp:
        assert resp.status == 200
        resp_json = await resp.json()
        assert_msg(resp_json, 'CREATE_DATASET')

    print("PUT request done, waiting for websocket response")

    # same msg via ws:
    msg = json.loads(await ws.recv())
    assert_msg(msg, 'CREATE_DATASET')

    return ds_uuid, ds_url


async def create_analysis(ws, http_client, base_url, ds_uuid, ca_uuid, details=None, token=None):
    analysis_uuid = _get_uuid_str()
    analysis_url = "{}/api/compoundAnalyses/{}/analyses/{}/".format(
        base_url, ca_uuid, analysis_uuid
    )
    analysis_url = add_token(analysis_url, token)
    if details is None:
        details = {
            "analysisType": "SUM_FRAMES",
            "parameters": {}
        }
    else:
        assert "analysisType" in details
        assert "parameters" in details
    analysis_data = {
        "dataset": ds_uuid,
        "details": details,
    }
    print("before PUT")
    async with http_client.put(analysis_url, json=analysis_data) as resp:
        print(await resp.text())
        assert resp.status == 200
        resp_json = await resp.json()
        assert resp_json['status'] == "ok"
    print("after PUT")

    msg = json.loads(await ws.recv())
    assert_msg(msg, 'ANALYSIS_CREATED')
    assert msg['dataset'] == ds_uuid
    assert msg['analysis'] == analysis_uuid
    assert msg['details'] == details

    return analysis_uuid, analysis_url


async def update_analysis(
    ws, http_client, base_url, ds_uuid, ca_uuid, analysis_uuid, details, token=None,
):
    analysis_url = "{}/api/compoundAnalyses/{}/analyses/{}/".format(
        base_url, ca_uuid, analysis_uuid
    )
    analysis_url = add_token(analysis_url, token)
    assert "analysisType" in details
    assert "parameters" in details
    analysis_data = {
        "dataset": ds_uuid,
        "details": details,
    }
    async with http_client.put(analysis_url, json=analysis_data) as resp:
        print(await resp.text())
        assert resp.status == 200
        resp_json = await resp.json()
        assert resp_json['status'] == "ok"

    msg = json.loads(await ws.recv())
    assert_msg(msg, 'ANALYSIS_UPDATED')
    assert msg['dataset'] == ds_uuid
    assert msg['analysis'] == analysis_uuid
    assert msg['details'] == details


async def create_update_compound_analysis(
    ws, http_client, base_url, ds_uuid, details=None, ca_uuid=None, token=None,
):
    if ca_uuid is None:
        ca_uuid = _get_uuid_str()
        creating = True
    else:
        creating = False
    ca_url = f"{base_url}/api/compoundAnalyses/{ca_uuid}/"
    ca_url = add_token(ca_url, token)

    if details is None:
        details = {
            "mainType": "APPLY_RING_MASK",
            "analyses": [],
        }
    else:
        assert "analyses" in details
        assert "mainType" in details

    ca_data = {
        "dataset": ds_uuid,
        "details": details,
    }

    print("dispatching PUT request...")

    async with http_client.put(ca_url, json=ca_data) as resp:
        print(await resp.text())

        assert resp.status == 200
        resp_json = await resp.json()
        assert resp_json['status'] == "ok"

    msg = json.loads(await ws.recv())
    if creating:
        assert_msg(msg, 'COMPOUND_ANALYSIS_CREATED')
    else:
        assert_msg(msg, 'COMPOUND_ANALYSIS_UPDATED')
    assert msg['dataset'] == ds_uuid
    assert msg['compoundAnalysis'] == ca_uuid
    assert msg['details'] == details

    return ca_uuid, ca_url


async def create_job_for_analysis(ws, http_client, base_url, analysis_uuid, token=None):
    job_uuid = _get_uuid_str()
    job_url = f"{base_url}/api/jobs/{job_uuid}/"
    job_data = {
        "job": {
            "analysis": analysis_uuid,
        }
    }
    job_url = add_token(job_url, token)
    async with http_client.put(job_url, json=job_data) as resp:
        print(await resp.text())
        assert resp.status == 200
        resp_json = await resp.json()
        assert resp_json['status'] == "ok"

    msg = json.loads(await ws.recv())
    assert_msg(msg, 'JOB_STARTED')
    assert msg['job'] == job_uuid
    assert msg['analysis'] == analysis_uuid
    assert msg['details']['id'] == job_uuid

    return job_uuid, job_url


async def call_ca_rpc(http_client, base_url, ca_uuid, rpc_name, token=None):
    rpc_url = f"{base_url}/api/compoundAnalyses/{ca_uuid}/rpc/{rpc_name}/"
    rpc_url = add_token(rpc_url, token)

    print(f"call CA before PUT {rpc_url}")

    async with http_client.put(rpc_url) as resp:
        print("call CA getting the text response:")
        print(await resp.text())
        resp_json = await resp.json()
        return resp_json



import asyncio
import signal
import time
import sys

import pytest


@pytest.fixture(scope="module")
def event_loop_policy(request):
    if sys.platform.startswith("win"):
        return asyncio.WindowsProactorEventLoopPolicy()
    return asyncio.get_event_loop_policy()  # TODO: windows fixes


@pytest.mark.asyncio
async def test_libertem_server_cli_startup(http_client):
    CTRL_C = signal.SIGINT
    # make sure we can start `libertem-server` and stop it again using ctrl+c
    # this is kind of a smoke test, which should cover the main cli functions.
    p = await asyncio.create_subprocess_exec(
        sys.executable, '-m', 'libertem.web.cli', '--no-browser',
        stderr=asyncio.subprocess.PIPE,
    )
    # total deadline, basically how long it takes to import all the dependencies
    # and start the web API
    # (no executor startup is done here)
    deadline = time.monotonic() + 15
    while True:
        if time.monotonic() > deadline:
            assert False, 'timeout'
        line = await asyncio.wait_for(p.stderr.readline(), 5)
        if not line:  # EOF
            assert False, 'subprocess is dead'
        line = line.decode("utf8")
        print('Line:', line, end='')
        if 'LiberTEM listening on' in line:
            break

    async def _debug():
        while True:
            line = await asyncio.wait_for(p.stderr.readline(), 5)
            if not line:  # EOF
                return
            line = line.decode("utf8")
            print('Line@_debug:', line, end='')

    debug_task = asyncio.ensure_future(_debug())

    try:
        addr = line[line.find("http"):].strip()
        async with http_client.get(addr, timeout=30.0) as response:
            assert response.status == 200, "Failed to GET from libertem-server"

        # windows likes to be special, so we just kill the subprocess instead:
        if sys.platform.startswith("win"):
            if p.returncode is None:
                p.terminate()
                await asyncio.sleep(0.2)
                await asyncio.wait_for(p.wait(), 1)
            if p.returncode is None:
                p.kill()
                await asyncio.wait_for(p.wait(), 1)
        else:
            # now, let's kill the subprocess:
            # ctrl+c twice should do the job:
            p.send_signal(CTRL_C)
            await asyncio.sleep(0.5)
            if p.returncode is None:
                p.send_signal(CTRL_C)

            # wait for the process to stop, but max. 1 second:
            ret = await asyncio.wait_for(p.wait(), 1)

            # in unix, minus signal number is returned if the process was killed:
            assert ret in (0, -CTRL_C)
    except Exception:
        if p.returncode is None:
            p.terminate()
            await asyncio.sleep(0.2)
        if p.returncode is None:
            p.kill()
        await asyncio.wait_for(p.wait(), 1)
        raise
    finally:
        debug_task.cancel()



import math
from typing import Any, Optional, overload
from collections.abc import Generator, Sequence

import numpy as np

from libertem.common.math import prod, count_nonzero
from libertem.common.shape import Shape, ShapeLike


class SliceUsageError(ValueError):
    """
    Raised when a Slice is incorrectly instantiated or used
    """


class Slice:
    """
    A n-dimensional slice, defined by origin and shape

    Parameters
    ----------
    origin : tuple of int
        global "top-left" coordinates of this slice
    shape : Shape instance
        the size of this slice
    """

    __slots__ = ["origin", "shape"]

    def __init__(self, origin: Sequence[int], shape: Shape):
        self.origin = tuple(origin)
        self.shape = shape
        if len(self.origin) != len(self.shape):
            raise SliceUsageError(
                ("cannot build slice with dimensionality of shape/origin mismatch (%d vs %d); "
                 "origin=%r, shape=%r") % (
                     len(self.origin), len(self.shape), self.origin, self.shape,
                 )
            )
        if not isinstance(shape, Shape):
            raise SliceUsageError("please use libertem.common.Shape instance as shape parameter")

    def __repr__(self) -> str:
        return f"<Slice origin={self.origin!r} shape={self.shape!r}>"

    def __hash__(self) -> int:
        # enables using a Slice as a key in dict, an item in sets etc.
        # in this case important for use as cache key for our mask container
        return hash((self.origin, tuple(self.shape)))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Slice) and (
            self.shape == other.shape and self.origin == other.origin
        )

    @classmethod
    def from_shape(cls, shape: Sequence[int], sig_dims: int) -> "Slice":
        """
        Construct a `Slice` at zero-origin from `shape` and `sig_dims`.
        """
        return Slice(
            origin=(0,) * len(shape),
            shape=Shape(shape, sig_dims=sig_dims),
        )

    def intersection_with(self, other: "Slice") -> "Slice":
        """
        Calculate the intersection between this slice and `other`. May result in
        dimensions that are zero, which means that there is no intersection.

        Returns
        -------
        slice : Slice
            the intersection between this and the other slice
        """
        if len(self.origin) != len(other.origin):
            raise SliceUsageError(
                ("cannot intersect slices with different dimensionality (%s vs %s); "
                 "self.origin=%r, other.origin=%r") % (
                    len(self.origin), len(other.origin), self.origin, other.origin,
                )
            )
        if self.shape.sig.dims != other.shape.sig.dims:
            raise SliceUsageError(
                "cannot intersect slices with different signal dimensionality ({} vs {})".format(
                    self.shape.sig.dims, other.shape.sig.dims
                )
            )
        new_origin = tuple(
            max(o1, o2)
            for (o1, o2) in zip(self.origin, other.origin)
        )
        new_shape = [
            min(
                (o1 + s1) - no,
                (o2 + s2) - no,
            )
            for (o1, o2, no, s1, s2) in zip(
                    self.origin, other.origin, new_origin, self.shape, other.shape
            )
        ]
        new_shape = [max(0, s) for s in new_shape]
        result = Slice(
            origin=new_origin,
            shape=Shape(new_shape, sig_dims=self.shape.sig.dims),
        )
        return result

    def is_null(self) -> bool:
        """
        If any part of our shape is zero, this slice doesn't span any data and is null / empty.
        """
        return any(s == 0 for s in self.shape)

    def shift(self, other: "Slice") -> "Slice":
        """
        make a new ``Slice`` with origin relative to ``other.origin``
        and the same shape as this ``Slice``

        useful for translating to the local coordinate system of ``other``
        """
        if len(self.origin) != len(other.origin):
            raise SliceUsageError(
                "cannot shift slices with different "
                f"dimensionality ({self.origin} vs {other.origin})"
            )
        return Slice(origin=tuple(our_coord - their_coord
                                  for (our_coord, their_coord) in zip(self.origin, other.origin)),
                     shape=self.shape)

    def shift_by(self, offset: Sequence[int]) -> "Slice":
        """
        Return a new slice with the origin moved by the supplied offset
        and the same shape
        """
        if len(self.origin) != len(offset):
            raise SliceUsageError(
                "cannot shift slices with different "
                f"dimensionality ({self.origin} vs {offset})"
            )
        return Slice(
            origin=tuple(
                our_coord + off
                for (our_coord, off)
                in zip(self.origin, offset)
            ),
            shape=self.shape,
        )

    @overload
    def get(
        self,
        arr: None = None,
        sig_only: bool = False,
        nav_only: bool = False
    ) -> tuple[slice, ...]: ...

    @overload
    def get(
        self,
        arr: np.ndarray,
        sig_only: bool = False,
        nav_only: bool = False
    ) -> np.ndarray: ...

    def get(
        self,
        arr: Optional[np.ndarray] = None,
        sig_only: bool = False,
        nav_only: bool = False
    ):
        """
        Get a standard python tuple-of-slice-object which can be used
        to slice any compatible numpy.ndarray

        Parameters
        ----------
        arr
            something implementing the slice interface. if given, returns arr[slice]
        sig_only : bool
            get a signal-only slice for frames/masks
        nav_only : bool
            get a nav-only slice, for example for indexing something that is shaped like
            the navigation dimensions of this Slice.

        Returns
        -------
        tuple of slice objects
            returns standard python slices computed from
            our origin+shape model or arr indexed with this slicing
            if arr is given


        Examples
        --------
        >>> import numpy as np
        >>> from libertem.common import Slice, Shape
        >>> s = Slice(shape=Shape((16, 16, 4, 4), sig_dims=2), origin=(0, 0, 12, 12))
        >>> data = np.ones((16, 16))
        >>> data[s.get(sig_only=True)]
        array([[1., 1., 1., 1.],
               [1., 1., 1., 1.],
               [1., 1., 1., 1.],
               [1., 1., 1., 1.]])
        """
        if sig_only:
            o, s = self.origin, self.shape
            slice_ = tuple(
                slice(o[i], (o[i] + s[i]))
                for i in range(s.nav.dims, s.sig.dims + s.nav.dims)
            )
        elif nav_only:
            o, s = self.origin, self.shape
            slice_ = tuple(
                slice(o[i], (o[i] + s[i]))
                for i in range(s.nav.dims)
            )
        else:
            slice_ = self._get()
        if arr is not None:
            if sig_only:
                # Skip the supposed nav dimensions of the data
                return arr[(Ellipsis, ) + slice_]
            else:
                # for nav_only, we return the full remaining dimensions anyway
                # if arr has more dimensions than the slice
                return arr[slice_]
        else:
            return slice_

    def _get(self):
        """
        Direct conversion from Slice to tuple(slice, ...) without options
        """
        return tuple(slice(o, (o + s)) for (o, s) in zip(self.origin, self.shape))

    def discard_nav(self) -> "Slice":
        """
        returns a copy with the origin/shape zeroed in the nav dimensions

        this is used to create uniform cache keys
        """
        o, s, sig_dims = self._discard_nav_key()
        new_shape = Shape(s, sig_dims=sig_dims)
        return Slice(origin=o, shape=new_shape)

    def _discard_nav_key(self) -> tuple[tuple[int, ...], tuple[int, ...], int]:
        """
        Construct a hashable tuple of the Slice with a zero-length nav dimensions

        Functions as discard_nav but avoids Shape and Slice constructor overheads
        """
        o, s = self.origin, self.shape
        nav_dims = s.nav_dims
        zero_nav = (0,) * nav_dims
        return (zero_nav + o[nav_dims:], zero_nav + s._sig_shape, s.sig_dims)

    def subslices(self, shape: ShapeLike) -> Generator["Slice", None, None]:
        """
        Generator for all subslices of this slice with dimensions
        specified by ``shape``.

        Parameters
        ----------
        shape : tuple of int or Shape
            the shape of each sub-slice

        Yields
        ------
        Slice
            all subslices, in fast-access order
        """
        # example: self.shape=(3, 1, 1, 1), subslice shape=(2, 1, 1, 1)
        # math.ceil(3/2) = math.ceil(1.5) = 2 -> we need two subslices across the y dimension
        shape = Shape(shape, sig_dims=self.shape.sig.dims)
        if self.shape.dims != shape.dims:
            raise SliceUsageError(
                ("cannot create subslices with different dimensionality (%d vs %d); "
                 "self.shape=%r, shape=%r") % (
                    self.shape.dims, shape.dims, self.shape, shape,
                )
            )
        ni = tuple(math.ceil(s1 / s)
                   for (s1, s) in zip(self.shape, shape))

        def _make_slice(origin: tuple[int, ...], new_shape: Shape) -> Slice:
            sig_dims = new_shape.sig.dims
            # this makes sure that the border tiles have the correct shape set
            new_shape_tuple = tuple(
                min(ns, so + s - o)
                for (ns, so, s, o) in zip(new_shape, self.origin, self.shape, origin)
            )
            new_shape = Shape(new_shape_tuple, sig_dims=sig_dims)
            for x in new_shape_tuple:
                assert x > 0, \
                    "invalid shape: {!r} while subslicing {!r} with {!r} (origin={!r})".format(
                        new_shape, self.shape, shape, origin
                    )
            return Slice(
                origin=origin,
                shape=new_shape,
            )

        return (
            _make_slice(origin=tuple(
                o + i * s
                for (o, i, s) in zip(self.origin, indexes, shape)
            ), new_shape=Shape(tuple(shape), sig_dims=self.shape.sig.dims))

            for indexes in np.ndindex(ni)
        )

    @property
    def nav(self) -> "Slice":
        """
        Returns a new Slice, with sig_dims=0, limited to the nav part
        """
        return Slice(
            origin=self.origin[:self.shape.nav.dims],
            shape=self.shape.nav,
        )

    @property
    def sig(self) -> "Slice":
        """
        Returns a new Slice, limited to the sig part
        """
        return Slice(
            origin=self.origin[self.shape.nav.dims:],
            shape=self.shape.sig,
        )

    def flatten_nav(self, containing_shape: ShapeLike) -> "Slice":
        sig_dims = self.shape.sig.dims
        nav_dims = self.shape.dims - sig_dims
        containing_shape = tuple(containing_shape)[:nav_dims]
        origin = self.origin[:nav_dims]

        # validation for the nav_shape:
        # what are the preconditions that allow flattening?
        #
        # - nav part of the shape: must be in the form of:
        #
        #   (1,  1,  ...,  N, M, M, ...)
        #
        #   where N<=M and M is the corresponding part of
        #   the shape of the dataset.
        #
        # - the origin must match the shape in the following way:
        #
        #   (o1, o2, ..., oi, 0, 0, ...)
        #
        #   where all oj are arbitraty (but in bounds)
        #
        state = 0
        for cs, s, o in zip(containing_shape, self.shape.nav, origin):
            if state == 0:
                if s != 1:
                    state = 1
                    assert s <= cs, "invalid nav_shape #1"
            elif state == 1:
                assert s == cs, "invalid nav_shape #2"
                assert o == 0, "invalid origin"

        nav_origin = np.ravel_multi_index(
            origin,
            containing_shape
        )
        nav_shape = prod(self.shape.nav)
        return Slice(
            origin=(nav_origin,) + self.origin[nav_dims:],
            shape=Shape((nav_shape,) + tuple(self.shape.sig), sig_dims=sig_dims)
        )

    def adjust_for_roi(self, roi: Optional[np.ndarray]) -> "Slice":
        """
        Make a new slice that has origin and shape modified according to `roi`.
        """
        if roi is None:
            return self
        roi = roi.reshape(-1)
        assert self.shape.nav.dims == 1
        s_o = self.origin[0]
        s_s = self.shape[0]
        # We need to find how many 1s there are for all previous partitions, to know
        # the origin; then we count how many 1s there are in our partition
        # to find our shape.
        origin = count_nonzero(roi[:s_o])
        shape = count_nonzero(roi[s_o:s_o + s_s])
        sig_dims = self.shape.sig.dims
        return Slice(
            origin=(origin,) + self.origin[-sig_dims:],
            shape=Shape((shape,) + tuple(self.shape.sig), sig_dims=sig_dims),
        )

    def clip_to(self, shape: Shape):
        other_slice = Slice((0,) * shape.dims, shape)
        return self.intersection_with(other_slice)

    def __getstate__(self) -> dict[str, Any]:
        return {
            k: getattr(self, k)
            for k in self.__slots__
        }

    def __setstate__(self, state: dict[str, Any]) -> None:
        for k, v in state.items():
            setattr(self, k, v)



from typing_extensions import Literal
from libertem.common.math import prod
import numpy as np

from libertem.common.udf import UDFMethod
from libertem.udf import UDF, UDFMeta
from libertem.common.buffers import AuxBufferWrapper
from libertem.common.container import MaskContainer
from libertem.common.numba import rmatmul


class ApplyMasksEngine:
    def __init__(self, masks: MaskContainer, meta: UDFMeta, use_torch: bool):
        self.masks = masks
        self.meta = meta

        try:
            import torch
        except ImportError:
            torch = None

        torch_incompatible = (
            torch is None
            or self.meta.input_dtype.kind != 'f'
            or self.meta.input_dtype != self.masks.dtype
            or self.meta.device_class != 'cpu'
            or self.meta.array_backend != UDF.BACKEND_NUMPY
            or self.masks.use_sparse
        )

        self.needs_transpose = True
        if use_torch and (not torch_incompatible):
            self.process_flat = self._process_flat_torch
        elif (
            self.meta.array_backend == UDF.BACKEND_NUMPY
            and isinstance(self.masks.use_sparse, str)
            and self.masks.use_sparse.startswith('scipy.sparse')
        ):
            # Due to https://github.com/scipy/scipy/issues/13211
            self.process_flat = self._process_flat_spsp
        elif (
            self.meta.array_backend in (
                UDF.BACKEND_SCIPY_COO,
                UDF.BACKEND_SCIPY_CSR,
                UDF.BACKEND_SCIPY_CSC
            ) and isinstance(self.masks.use_sparse, str)
            and self.masks.use_sparse.startswith('sparse.pydata')
        ):
            self.process_flat = self._process_flat_sparsepyd
            self.needs_transpose = False
        else:
            self.process_flat = self._process_flat_standard

    def _get_masks(self):
        return self.masks.get_for_sig_slice(
            self.meta.sig_slice, transpose=self.needs_transpose
        )

    def _process_flat_torch(self, flat_tile, masks):
        import torch
        # CuPy back-end disables torch in get_task_data
        # FIXME use GPU torch with CuPy array?
        return torch.mm(
            torch.from_numpy(flat_tile),
            torch.from_numpy(masks),
        ).numpy()

    def _process_flat_spsp(self, flat_tile, masks):
        return rmatmul(flat_tile, masks)

    def _process_flat_sparsepyd(self, flat_tile, masks):
        # Make sure the sparse.pydata mask comes first
        # to choose the right multiplication method
        return (masks @ flat_tile.T).T

    def _process_flat_standard(self, flat_tile, masks):
        return flat_tile @ masks

    def process_tile(self, tile):
        flat_shape = (tile.shape[0], prod(tile.shape[1:]))
        # Avoid reshape since older versions of scipy.sparse don't support it
        flat_data = tile.reshape(flat_shape) if tile.shape != flat_shape else tile
        return self.process_flat(flat_data, self._get_masks())

    def process_frame_shifted(self, frame, shifts: tuple[int, ...]):
        sig_shape = self.meta.dataset_shape.sig
        masks = self._get_masks()
        num_masks = len(self.masks)
        shifted_slice = self.meta.sig_slice.shift_by(shifts)
        inverse_shifted_slice = self.meta.sig_slice.shift_by(-1 * shifts)
        left = self.meta.sig_slice.intersection_with(shifted_slice)
        right = self.meta.sig_slice.intersection_with(inverse_shifted_slice)
        if left.is_null():
            # Zero overlap after shifts, shortcut return
            return np.zeros((num_masks,), dtype=np.float32)
        mask_slice = right.get()
        if self.needs_transpose:
            # expects masks in shape (sig_size, num_masks)
            mask_slice = mask_slice + (slice(None), )
            masks = masks.reshape((*sig_shape, num_masks))
            final_mask_shape = (-1, num_masks)
        else:
            # expects masks in shape (num_masks, sig_size)
            # NOTE unexpectedly don't need to reverse sig_shape or mask_slice ?
            masks = masks.reshape((num_masks, *sig_shape))
            mask_slice = (slice(None), ) + mask_slice
            final_mask_shape = (num_masks, -1)

        sliced_masks = masks[mask_slice].reshape(final_mask_shape)
        # shift slicing requires sig_shape frames
        # sparse array backend can provide flat frame
        frame = frame.reshape(sig_shape)
        try:
            data = left.get(frame)
        except TypeError as e:
            # frame is in a form which doesn't support slicing
            # the only recognized case is scipy.sparse.coo
            if not hasattr(frame, 'getformat'):
                raise e  # pragma: no cover
            assert frame.getformat() == 'coo'
            frame = frame.tocsr()
            data = left.get(frame)
        flat_data = data.reshape((1, -1))
        return self.process_flat(flat_data, sliced_masks).reshape((num_masks,))


class ApplyMasksUDF(UDF):
    '''
    Apply masks to signals/frames in the dataset. This can not only be used to integrate
    over regions with a binary mask - the integration can be weighted by using
    float or complex valued masks.

    The result will be returned in a single sig-shaped buffer called intensity.
    Its shape will be :code:`(*nav_shape, len(masks))`.

    Parameters
    ----------

    mask_factories : Union[Callable[[], array_like], Iterable[Callable[[], array_like]]]
        Function or list of functions that take no arguments and create masks. The returned
        masks can be
        numpy arrays, scipy.sparse or sparse https://sparse.pydata.org/ matrices. The mask
        factories should not reference large objects because they can create significant
        overheads when they are pickled and unpickled. Each factory function should, when
        called, return a numpy array with the same shape as frames in the dataset
        (so dataset.shape.sig).
    use_torch : bool, optional
        Use pytorch back-end if available. Default True
    use_sparse : Union[None, False, True, 'scipy.sparse', 'scipy.sparse.csc', \
            'sparse.pydata'], optional
        Which sparse back-end to use.
        * None (default): Where possible use sparse matrix multiplication if all factory \
            functions return a sparse mask, otherwise convert all masks to dense matrices \
            and use dense matrix multiplication
        * True: Convert all masks to sparse matrices and use default sparse back-end.
        * False: Convert all masks to dense matrices.
        * 'scipy.sparse': Use scipy.sparse.csr_matrix (default sparse)
        * 'scipy.sparse.csc': Use scipy.sparse.csc_matrix
        * 'sparse.pydata': Use sparse.pydata COO matrix
    mask_count : int, optional
        Specify the number of masks if a single factory function is used so that the
        number of masks can be determined without calling the factory function.
    mask_dtype : numpy.dtype, optional
        Specify the dtype of the masks so that mask dtype
        can be determined without calling the mask factory functions. This can be used to
        override the mask dtype in the result dtype determination. As an example, setting
        this to np.float32 means that masks of type float64 will not switch the calculation
        and result dtype to float64 or complex128.
    preferred_dtype : numpy.dtype, optional
        Let :meth:`get_preferred_input_dtype` return the specified type instead of the
        default `float32`. This can perform the calculation with integer types if both input
        data and mask data are compatible with this.
    backends : Iterable containing strings "numpy" and/or "cupy", or None
        Control which back-ends are used. Default is numpy and cupy
    shifts : Union[Tuple[int, int], AuxBufferWrapper], optional
        (Y/X)-shifts to apply to all masks before multiplying with each frame. Can be either a
        length-2 array-like for a constant :code:`(Y, X)` shift, or an
        :class:`~libertem.common.buffers.AuxBufferWrapper` of
        :code:`(kind='nav', extra_shape=(2,), dtype=int)` defining a per-frame shift to apply.

        A positive y-shift moves the mask 'down' relative to the frame, while a positive
        x-shift moves the mask 'right' relative to the frame. Elements of the mask and frame
        which do not overlap after the shift are discarded. A shift resulting in no overlap
        at all will return a sum of :code:`0.` for that frame.

        .. note::
            Float shift values are cast to integers internally; round values before
            passing the shifts argument to better control the exact shifts used.

        .. note::
            The :code:`shifts` parameter requires frame-by-frame processing to function, and so
            adds a performance penalty compared to unshifted mask application. If applying a
            constant shift it may be worthwhile to manually create a new, pre-shifted mask
            rather than relying on this feature.

            Shifting is also currently incompatible with :code:`scipy.sparse` masks. If sparse
            processing is required then where possible :code:`scipy.sparse` masks are converted to
            :code:`sparse.pydata` equivalents. A consequence of this is that sparse processing
            is not yet supported through CuPy when shifts are enabled, as :code:`sparse.pydata`
            has no current CuPy implementation. If sparse masks are supplied on a CuPy backend
            when :code:`use_sparse=None` (the default) they will be densified to allow the
            calculation to take place.

    Examples
    --------

    >>> dataset.shape
    (16, 16, 32, 32)
    >>> def my_masks():
    ...     return [np.ones((32, 32)), np.zeros((32, 32))]
    >>> udf = ApplyMasksUDF(mask_factories=my_masks)
    >>> res = ctx.run_udf(dataset=dataset, udf=udf)['intensity']
    >>> res.data.shape
    (16, 16, 2)
    >>> np.allclose(res.data[..., 1], 0)  # same order as in the mask factory
    True

    Mask factories can also return all masks as a single array, stacked on the first axis:

    >>> def my_masks_2():
    ...     masks = np.zeros((2, 32, 32))
    ...     masks[1, ...] = 1
    ...     return masks
    >>> udf = ApplyMasksUDF(mask_factories=my_masks_2)
    >>> res_2 = ctx.run_udf(dataset=dataset, udf=udf)['intensity']
    >>> np.allclose(res_2.data, res.data)
    True

    Masks can be shifted relative to the data using the :code:`shifts` parameter,
    this can either be a constant shift for all frames:

    >>> udf = ApplyMasksUDF(mask_factories=my_masks, shifts=(2, -5))
    >>> res_shift_constant = ctx.run_udf(dataset=dataset, udf=udf)['intensity']

    or a per-frame shift supplied using an :class:`~libertem.common.buffers.AuxBufferWrapper`
    created using :meth:`~libertem.udf.base.UDF.aux_data`:

    >>> udf = ApplyMasksUDF(
    ...         mask_factories=my_masks,
    ...         shifts=ApplyMasksUDF.aux_data(
    ...             np.random.randint(-8, 8, size=(16, 16, 2)).ravel(),
    ...             kind='nav',
    ...             extra_shape=(2,),
    ...             dtype=int,
    ...         )
    ...     )
    >>> res_shift_aux = ctx.run_udf(dataset=dataset, udf=udf)['intensity']

    .. versionadded:: 0.4.0

    .. versionchanged:: 0.13.0
        Added the :code:`shifts` parameter
    '''
    def __init__(self, mask_factories, use_torch=True, use_sparse=None, mask_count=None,
                mask_dtype=None, preferred_dtype=None, backends=None, shifts=None, **kwargs):

        _backends = backends
        if backends is None:
            backends = self.BACKEND_ALL
        backends = tuple(b for b in backends if b in self.BACKEND_ALL)

        if shifts is not None:
            if isinstance(use_sparse, str) and use_sparse.startswith('scipy.sparse'):
                # This is 'unsupported' because we need to slice the mask stack
                # in the signal dimensions to shift it, and sig is flattened
                # to give to 2D matrix in scipy.sparse
                raise ValueError(
                    f'Sparse backend {use_sparse} not supported for '
                    'shifts, use sparse.pydata instead.'
                )
            if not isinstance(shifts, AuxBufferWrapper):
                shifts = np.asarray(shifts)

            backends = tuple(
                b for b in backends
                if b not in (
                    # Here we are doing frame-by-frame processing, so we can
                    # accept scipy.sparse-style frames, however we need to
                    # perform a reshape into sig-shaped frames which normally
                    # casts the frame into coo_matrix form, which then
                    # has to be re-cast into csr_matrix form to be sliced (i.e. shifted)
                    self.BACKEND_SCIPY_COO,  # cannot be sliced
                    self.BACKEND_CUPY_SCIPY_COO,  # cannot be sliced
                )
            )

        if len(backends) == 0:
            raise ValueError(f'No compatible backend found in {_backends}')

        self._mask_container = None

        super().__init__(
            mask_factories=mask_factories,
            use_torch=use_torch,
            use_sparse=use_sparse,
            mask_count=mask_count,
            mask_dtype=mask_dtype,
            preferred_dtype=preferred_dtype,
            backends=backends,
            shifts=shifts,
            **kwargs
        )

    def get_preferred_input_dtype(self):
        ''
        if self.params.preferred_dtype is None:
            return super().get_preferred_input_dtype()
        else:
            return self.params.preferred_dtype

    def get_mask_dtype(self):
        if self.params.mask_dtype is None:
            return self.masks.dtype
        else:
            return self.params.mask_dtype

    def get_mask_count(self):
        if self.params.mask_count is None:
            return len(self.masks)
        else:
            return self.params.mask_count

    @property
    def masks(self):
        if self._mask_container is None:
            self._mask_container = self._make_mask_container()
        return self._mask_container

    def _make_mask_container(self):
        p = self.params
        if self.meta.array_backend in self.CUPY_BACKENDS:
            backend = self.BACKEND_CUPY
        else:
            backend = self.BACKEND_NUMPY
        # In the default case defer to default kwarg on MaskContainer
        default_sparse = {}
        if p.shifts is None:
            default_sparse['default_sparse'] = 'scipy.sparse'
        else:
            default_sparse['default_sparse'] = 'sparse.pydata'
        return MaskContainer(
            p.mask_factories, dtype=p.mask_dtype, use_sparse=p.use_sparse, count=p.mask_count,
            backend=backend, **default_sparse,
        )

    def get_task_data(self):
        ''
        engine = ApplyMasksEngine(self.masks, self.meta, self.params.use_torch)
        return {
            'engine': engine,
        }

    def get_result_buffers(self):
        ''
        dtype = np.result_type(self.meta.input_dtype, self.get_mask_dtype())
        count = self.get_mask_count()
        return {
            'intensity': self.buffer(
                kind='nav', extra_shape=(count, ), dtype=dtype, where='device'
            )
        }

    def get_backends(self):
        ''
        return self.params.backends

    def get_method(self) -> Literal[UDFMethod.FRAME, UDFMethod.TILE]:
        """
        :meta private:
        """
        if self.params.get('shifts') is not None:
            return UDFMethod.FRAME
        else:
            return UDFMethod.TILE

    def process_tile(self, tile):
        """
        Used for simple mask application, without shifts

        :meta private:
        """
        self.results.intensity[:] += self.forbuf(
            self.task_data.engine.process_tile(tile),
            self.results.intensity,
        )

    def process_frame(self, frame):
        """
        Apply shifted masks to a frame

        :meta private:
        """
        shifts = self.params.shifts.astype(int)
        self.results.intensity[:] += self.forbuf(
            self.task_data.engine.process_frame_shifted(frame, shifts),
            self.results.intensity,
        )



import os
import glob
import pickle
import shutil
import importlib

import scipy.sparse
import pytest
import numpy as np
import numba

from libertem.api import Context
from libertem.common.numba import rmatmul, numba_dtypes
from libertem.common.numba.cache import _cached_njit_reg
from libertem.web.dataset import prime_numba_cache


@pytest.mark.with_numba
def test_rmatmul_csr():
    le = np.random.random((2, 3))
    ri = scipy.sparse.csr_matrix(np.random.random((3, 2)))
    assert np.allclose(rmatmul(le, ri), le @ ri)


@pytest.mark.with_numba
def test_rmatmul_csc():
    le = np.random.random((2, 3))
    ri = scipy.sparse.csr_matrix(np.random.random((3, 2)))
    assert np.allclose(rmatmul(le, ri), le @ ri)


def test_rmatmul_1():
    le = np.zeros((1, 2, 3))
    ri = scipy.sparse.csr_matrix(np.zeros((5, 6)))
    # 3D shape left
    with pytest.raises(ValueError):
        rmatmul(le, ri)


def test_rmatmul_2():
    le = np.zeros((2, 3))
    ri = np.zeros((4, 5, 6))
    # 3D shape right
    with pytest.raises(ValueError, ):
        rmatmul(le, ri)


def test_rmatmul_3():
    le = np.zeros((2, 3))
    ri = scipy.sparse.csr_matrix(np.zeros((5, 6)))
    # Shape mismatch
    with pytest.raises(ValueError):
        rmatmul(le, ri)


def test_rmatmul_4():
    le = np.zeros((2, 3))
    ri = np.zeros((3, 2))
    # Not a csc or csr matrix
    with pytest.raises(ValueError):
        rmatmul(le, ri)


def test_numba_cache(tmpdir_factory):
    src_dir = tmpdir_factory.mktemp('numba_cache')
    src_py_file = os.path.join(os.path.dirname(__file__), 'numba_cache.py')
    shutil.copy(src_py_file, src_dir)

    dst_py_file = os.path.join(src_dir, 'numba_cache.py')
    spec = importlib.util.spec_from_file_location('<dynamic>', dst_py_file)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)

    pycache_dir = os.path.join(src_dir, '__pycache__')

    nbis = glob.glob(pycache_dir + '/*.nbi')
    nbcs = glob.glob(pycache_dir + '/*.nbc')

    assert len(_cached_njit_reg) >= 1
    assert len(nbis) == 1
    assert len(nbcs) == 2

    with open(nbis[0], 'rb') as f:
        pickle.load(f)  # load and discard version
        data = f.read()
        cache_contents = pickle.loads(data)

    assert len(cache_contents[1].keys()) == 2


@pytest.mark.slow
def test_numba_prime(default_raw):
    prime_numba_cache(default_raw)


def test_numba_prime_hdf5_1(hdf5_ds_1):
    prime_numba_cache(hdf5_ds_1)


def test_numba_prime_hdf5_2(hdf5_ds_large_sig):
    prime_numba_cache(hdf5_ds_large_sig)


@numba.njit
def numba_sum(arr):
    return arr.sum()


@pytest.mark.parametrize(
    'dtype', numba_dtypes
)
def test_numba_dtype(dtype):
    arr = np.ones(42, dtype=dtype)
    res = numba_sum(arr)
    assert arr.dtype in numba_dtypes
    assert res == len(arr)


@numba.njit
def _issue_1432():
    np.array([1 for _ in range(1)])


@pytest.mark.slow
def test_issue_1432():
    from libertem.executor.dask import cluster_spec, DaskJobExecutor
    spec = cluster_spec(cpus=1, cudas=[], has_cupy=False, num_service=0)
    executor = DaskJobExecutor.make_local(spec=spec)
    with Context(executor=executor) as ctx:
        print(ctx.executor.run_function(_issue_1432))



from unittest import mock
import sys

import numpy as np
import pytest

from libertem.viz.gms import GMSLive2DPlot
from libertem.udf.sum import SumUDF
from libertem.udf.sumsigudf import SumSigUDF


def test_gms_smoke(monkeypatch, lt_ctx, default_raw):
    udf = SumUDF()
    monkeypatch.setitem(sys.modules, 'DigitalMicrograph', mock.MagicMock())
    monkeypatch.setattr(lt_ctx, 'plot_class', GMSLive2DPlot)
    lt_ctx.run_udf(dataset=default_raw, udf=udf, plots=True)
    sys.modules['DigitalMicrograph'].CreateImage.assert_called()


def test_gms_nodisplay(monkeypatch, lt_ctx, default_raw):
    udf = SumUDF()
    monkeypatch.setitem(sys.modules, 'DigitalMicrograph', mock.MagicMock())
    plots = [GMSLive2DPlot(dataset=default_raw, udf=udf)]
    with pytest.warns(UserWarning, match="Plot is not displayed, not plotting."):
        lt_ctx.run_udf(dataset=default_raw, udf=udf, plots=plots)


def test_empty(monkeypatch, lt_ctx, default_raw):
    udf = SumSigUDF()
    monkeypatch.setitem(sys.modules, 'DigitalMicrograph', mock.MagicMock())
    monkeypatch.setattr(lt_ctx, 'plot_class', GMSLive2DPlot)

    def extract(udf_result, damage):
        return udf_result['intensity'].data, np.zeros_like(damage)

    plot = GMSLive2DPlot(dataset=default_raw, udf=udf, channel=extract)

    with pytest.warns(UserWarning, match='^Plot is not displayed, not plotting'):
        lt_ctx.run_udf(dataset=default_raw, udf=udf, plots=[plot])



from libertem.io.corrections.detector import (
    correct, RepairDescriptor, RepairValueError, correct_dot_masks
)

__all__ = ['correct', 'RepairDescriptor', 'RepairValueError', 'correct_dot_masks']



import numpy as np
import pytest
from unittest.mock import Mock

from libertem import viz
from libertem.viz.base import Dummy2DPlot
from libertem.udf.base import NoOpUDF
from libertem.udf.sum import SumUDF
from libertem.udf.sumsigudf import SumSigUDF


def test_rgb_from_vector():
    rgb = viz.CMAP_CIRCULAR_DEFAULT.rgb_from_vector((0, 0, 0))  # center (grey)
    np.testing.assert_equal(rgb, np.asarray([127, 127, 127], dtype=np.uint8))
    rgb = viz.CMAP_CIRCULAR_DEFAULT.rgb_from_vector((0, 1, 0))  # up (green)
    np.testing.assert_equal(rgb, np.asarray([78, 173, 19], dtype=np.uint8))
    rgb = viz.CMAP_CIRCULAR_DEFAULT.rgb_from_vector((1, 0, 0))  # right (red)
    np.testing.assert_equal(rgb, np.asarray([230, 87, 64], dtype=np.uint8))
    rgb = viz.CMAP_CIRCULAR_DEFAULT.rgb_from_vector((0, -1, 0))  # down (purple)
    np.testing.assert_equal(rgb, np.asarray([177, 81, 234], dtype=np.uint8))
    rgb = viz.CMAP_CIRCULAR_DEFAULT.rgb_from_vector((-1, 0, 0))  # left (cyan)
    np.testing.assert_equal(rgb, np.asarray([24, 167, 191], dtype=np.uint8))


@pytest.mark.parametrize("log", [True, False])
def test_all_nan(log):
    data = np.full((16, 16), np.nan)
    viz.visualize_simple(data, logarithmic=log)


@pytest.mark.parametrize("log", [True, False])
def test_all_ones(log):
    data = np.ones((16, 16))
    viz.visualize_simple(data, logarithmic=log)


@pytest.mark.parametrize("log", [True, False])
def test_all_zeros(log):
    data = np.zeros((16, 16))
    viz.visualize_simple(data, logarithmic=log)


@pytest.mark.parametrize("log", [True, False])
def test_all_negative(log):
    data = np.full((16, 16), -1)
    viz.visualize_simple(data, logarithmic=log)


@pytest.mark.parametrize("log", [True, False])
def test_some_nonnegative(log):
    data = np.full((16, 16), -1)
    data[0] = 0
    data[1] = 1
    viz.visualize_simple(data, logarithmic=log)


def test_live_nochannels(default_raw):
    udf = NoOpUDF()
    with pytest.raises(ValueError):
        Dummy2DPlot(dataset=default_raw, udf=udf)


def test_live_plotupdate(lt_ctx, default_raw):
    udf = SumUDF()
    m = Mock()
    m.get_udf.return_value = udf
    plots = [m]
    lt_ctx.run_udf(dataset=default_raw, udf=udf, plots=plots)
    n_part = default_raw.get_num_partitions()
    print("Num partitions", n_part)
    print("Mock calls", m.new_data.mock_calls)
    # Otherwise test is meaningless, no intermediate updates
    assert n_part > 1
    # Partition updates plus final update
    assert len(m.new_data.mock_calls) == n_part + 1


@pytest.mark.parametrize('udf_cls', (SumUDF, SumSigUDF))
def test_live_autoextraction(lt_ctx, default_raw, udf_cls):
    udf = udf_cls()
    plots = [Dummy2DPlot(dataset=default_raw, udf=udf, channel=('intensity', np.abs))]
    lt_ctx.run_udf(dataset=default_raw, udf=udf, plots=plots)


def test_live_RGB(lt_ctx, default_raw):
    udf = SumSigUDF()

    def RGB_plot(udf_result, damage):
        data = udf_result['intensity'].data
        plot = viz.CMAP_CIRCULAR_DEFAULT.rgb_from_vector((data, 0, 0))
        return (plot, damage)

    plots = [Dummy2DPlot(dataset=default_raw, udf=udf, channel=RGB_plot)]
    lt_ctx.run_udf(dataset=default_raw, udf=udf, plots=plots)

    assert plots[0].data.shape == tuple(default_raw.shape.nav) + (3, )



from .exceptions import DataSetException
from .meta import DataSetMeta, PartitionStructure
from .roi import _roi_to_nd_indices
from .backend import IOBackend
from .backend_mmap import MMapBackend
from .backend_buffered import BufferedBackend
from .backend_direct import DirectBackend
from .dataset import DataSet, WritableDataSet
from .partition import Partition, BasePartition, WritablePartition
from .utils import FileTree
from .fileset import FileSet
from .file import File
from .tiling import (
    DataTile, default_get_read_ranges, make_get_read_ranges,
)
from .tiling_scheme import TilingScheme, Negotiator
from .decode import (
    Decoder, DtypeConversionDecoder, decode_swap_2, decode_swap_4,
)
from .coordinates import get_coordinates

__all__ = [
    'DataSetException', 'DataSetMeta', 'PartitionStructure',
    '_roi_to_nd_indices',
    'DataSet', 'WritableDataSet', 'Partition', 'WritablePartition', 'BasePartition',
    'DataTile', 'FileSet', 'File',
    'FileTree', 'TilingScheme',
    'default_get_read_ranges', 'make_get_read_ranges',
    'Decoder', 'DtypeConversionDecoder', 'Negotiator',
    'decode_swap_2', 'decode_swap_4', 'get_coordinates',
    'IOBackend', 'BufferedBackend', 'MMapBackend', 'DirectBackend',
]



from typing import (
    Any, Callable, Optional,
)
from collections.abc import Iterable, Generator


def default_unpackable() -> dict[type, Callable[[Iterable], Iterable[tuple[Any, Any]]]]:
    """
    Defaults for types which can be unpacked by the
    functions in this file, providing a mapping from
    type to a fn(instance) giving an iterable yielding
    (index, element) within the unpackable.
    """
    _unpackable_types = {list: lambda x: enumerate(x),
                        tuple: lambda x: enumerate(x),
                        dict: lambda x: x.items()}
    return _unpackable_types


def default_merge_fns() -> dict[type, Callable[[Iterable, Any, Any], None]]:
    """
    Default merge functions for rebuilding structures

    Returns a mapping from {type: callable} with callable signature:
        fn(structure, el, position)
    which inserts el into the structure at position
    """
    merge_fns = {list: lambda lis, el, pos: lis.append(el),
                dict: lambda dic, el, pos: dic.update({pos: el})}
    return merge_fns


class IgnoreClass:
    """
    A Null type used as a signal when None or False have meaning
    """
    pass


class StructDescriptor:
    def __init__(self, cls: type, *args: tuple, **kwargs: dict):
        """
        Container to hold a reference to a type and both the
        args and kwargs necessary to initialise it

        Could be replaced with a simple dict, namedtuple etc but being
        able to *args, **kwargs for variable numbers of parameters
        is quite useful
        """
        self.cls = cls
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        return f'{self.__class__.__name__}({self.cls}, {self.args}, {self.kwargs})'


def flatten_nested(el: Any,
                   unpackable_types: dict[type, Callable[[Iterable],
                                                         Iterable[tuple[Any, Any]]]] = None,
                   ignore_types: Optional[tuple[type, ...]] = None) -> list[Any]:
    """
    Recursively unpack the structure el while the type of el is in
    the mapping unpackable_types, which maps between the types that
    can be unpacked and a function fn(el) returning an iterable
    that gives tuples of (index, subelement),
    e.g. enumerate(['a', 'b']) => (0, 'a'), (1, 'b')

    ignore_types are types which this function should specifically ignore
    and not try to unpack, even if they are present in unpackable_types

    Returns a flat list containing all the elements of the
    (possibly nested) structure el
    """
    eltype = type(el)
    if unpackable_types is None:
        unpackable_types = default_unpackable()
    if ignore_types is None:
        ignore_types = (IgnoreClass,)
    flattened = []
    if eltype in unpackable_types.keys() and not isinstance(el, ignore_types):
        # If el is unpackable but empty
        if not el:
            flattened.append(IgnoreClass)
            return flattened
        iterable = unpackable_types[eltype](el)
        for _, _el in iterable:
            flattened.extend(flatten_nested(_el,
                                            unpackable_types=unpackable_types,
                                            ignore_types=ignore_types))
    else:
        flattened.append(el)
    return flattened


def build_mapping(el: Any,
                  unpackable_types: dict[type, Callable[[Iterable],
                                                        Iterable[tuple[Any, Any]]]] = None,
                  ignore_types: Optional[tuple[type, ...]] = None,
                  _pos: list[tuple[type, Any]] = None) -> list[list[tuple[type, Any]]]:
    """
    Recursively unpack the structure el and build a flat descriptor of its
    structure, such that it can be re-built

    The elements of the return list are essentially each a list of 'coordinates'
    that map from a position in the flattened version of el, to the position
    in the (possibly nested) original structure el

    Same arguments as flatten_nested except pos, which is only used by the
    function as a way to pass the current position down to the next
    level of the function calls
    """
    flat_mapping = []
    eltype = type(el)
    if unpackable_types is None:
        unpackable_types = default_unpackable()
    if ignore_types is None:
        ignore_types = (IgnoreClass,)
    if eltype in unpackable_types.keys() and not isinstance(el, ignore_types):
        # If el is unpackable but empty
        if not el:
            flat_mapping.append(_pos + [(eltype, IgnoreClass)])
            return flat_mapping
        iterable = unpackable_types[eltype](el)
        for __pos, _el in iterable:
            if _pos is None:
                _pos = []
            flat_mapping.extend(build_mapping(_el, _pos=_pos + [(eltype, __pos)],
                                              unpackable_types=unpackable_types,
                                              ignore_types=ignore_types))
    else:
        flat_mapping.append(_pos)
    return flat_mapping


def rebuild_nested(flat: list[Any],
                   flat_mapping: list[list[tuple[type, Any]]],
                   merge_functions=None):
    """
    Using the flattened version of a structure built by flatten_nested
    and the coordinates created by build_mapping, reconstruct the original
    nested structure

    merge_functions is a mapping from type: fn() with signature:
        fn(_nest, el, position)
    which inserts el into the structure _nest at position

    By default this function only knows how to rebuild a nest
    consisting of [list, dict, tuple], and in the tuple case actually
    reconstructs as list before casting to tuple at the end (to avoid
    immutability of tuples). In principle, by supplying extra merge_functions
    this function should be able to reconstruct other mutable iterables.

    This function works left-to-right in the list flat.
    Could perhaps be done better by building from deepest
    to shallowest across the set of elements in flat.
    """
    if merge_functions is None:
        merge_functions = default_merge_fns()
    nest = None
    for el, coords in zip(flat, flat_mapping):
        # Build the outer iterable of the structure
        if nest is None:
            # Case of a non-unpackable initial element
            if coords is None:
                return el
            nest_class = coords[0][0]
            # Hack tuples to list to avoid immutability problems
            if nest_class == tuple:
                nest_class = list
            nest = nest_class()
        nest = insert_at_pos(el, coords, nest, merge_functions)
    # Convert hacked lists into tuples, from deepest to shallowest
    nest = list_to_tuple(nest, flat_mapping)
    return nest


def pairwise(iterable: Iterable[Any]) -> Generator[tuple[Any, Any], None, None]:
    """
    Yield elements of iterable as tuples of overlapping pairs
    finally yielding (last_element, None)
    """
    for el in iterable:
        try:
            yield prior_el, el
        except NameError:
            pass
        prior_el = el
    yield prior_el, None


def insert_at_pos(el: Any,
                  coords: list[tuple[type, Any]],
                  nest: Iterable,
                  merge_fns: dict[type, Callable[[Iterable, Any, Any], None]]):
    """
    For the partially completed nested structure nest, insert the
    element el at the position given by coords

    If the position of el does not exist yet, build the structure
    from the top down until el can be inserted

    merge el into existing structures using a function from
    the mapping merge_fns[type(el)](_nest, el, position)

    tuples are treated as lists to allow appending, and are later
    converted to tuples once the nest is completed
    """
    _nest = nest
    for current_coord, next_coord in pairwise(coords):
        current_cls, current_pos = current_coord
        next_cls, next_pos = None, None
        if next_coord is not None:
            next_cls, next_pos = next_coord
        # Handle empty structures, these are marked with a special
        # coordinate (cls, IgnoreClass) and by definition are the
        # end of insertions for this set of coords
        if next_pos is IgnoreClass:
            merge_fns[type(_nest)](_nest, next_cls(), current_pos)
            return nest
        # Hack tuples to lists to avoid immutability problems
        if next_cls == tuple:
            next_cls = list
        if next_pos is None:
            merge_fns[type(_nest)](_nest, el, current_pos)
        else:
            try:
                _nest = _nest[current_pos]
            except KeyError:
                _nest[current_pos] = next_cls()
                _nest = _nest[current_pos]
            except IndexError:
                _nest.append(next_cls())
                _nest = _nest[current_pos]
    return nest


def find_tuples(flat_mapping: list[list[tuple[type, Any]]]) -> list[tuple[int, int]]:
    """
    Get the indexes in flat_mapping and depth in the coordinate
    where the coordinate specify the structure is of class tuple
    """
    return [(i, j) for i, coord in enumerate(flat_mapping)
            for j, _coord in enumerate(coord)
            if _coord[0] == tuple]


def set_as_tuple(nest, indices: list[Any]):
    """
    For a given sequence of indices to index into the completed
    nest, convert the structure at the final index in the sequence
    to a tuple type (if it is not already a tuple)
    """
    if (len(indices) > 1) and isinstance(nest[indices[0]], (dict, list)):
        set_as_tuple(nest[indices[0]], indices[1:])
    else:
        nest[indices[0]] = tuple(nest[indices[0]])


def list_to_tuple(nest, flat_mapping: list[list[tuple[type, Any]]]):
    """
    Convert any elements which are marked as tuples in flat_mapping
    but were constructed in nest as lists, back into tuples
    """
    tuple_positions = find_tuples(flat_mapping)
    deepest_first = reversed(sorted(tuple_positions, key=lambda x: x[1]))
    for coord_i, depth_j in deepest_first:
        indices = [c[1] for c in flat_mapping[coord_i][:depth_j]]
        if len(indices) > 0:
            set_as_tuple(nest, indices)
        else:
            nest = tuple(nest)
    return nest



import os
import time
from glob import glob
import concurrent.futures
import multiprocessing.pool
from typing import Any

import pytest
import distributed
import dask
import numpy as np
from libertem.common.shape import Shape
from libertem.common.slice import Slice

from libertem.utils.devices import detect
from libertem.executor.delayed import DelayedJobExecutor
from libertem.executor.dask import DaskJobExecutor
from libertem.executor.concurrent import ConcurrentJobExecutor
from libertem.executor.inline import InlineJobExecutor
from libertem.common.executor import (
    Environment, TaskCommHandler, TaskProtocol, WorkerQueue,
)
from libertem.api import Context
from libertem.udf.stddev import StdDevUDF
from libertem.udf.masks import ApplyMasksUDF
from libertem.common.exceptions import ExecutorSpecException
from sparseconverter import (
    BACKENDS, CUPY, CUPY_BACKENDS, CUPY_SCIPY_CSR, NUMPY, SCIPY_COO, SPARSE_COO
)

from utils import get_testdata_path


d = detect()
has_cupy = d['cudas'] and d['has_cupy']
has_gpus = len(d['cudas']) > 0


@pytest.fixture(
    scope='module',
    params=[
        "inline", "dask_executor", "dask_make_default", "dask_integration",
        "concurrent", "delayed", "pipelined",
    ]
)
def ctx(request, local_cluster_url):
    print(f'Make ctx with {request.param}')
    if request.param == 'inline':
        yield Context.make_with('inline')
    elif request.param == "dask_executor":
        dask_executor = DaskJobExecutor.connect(local_cluster_url)
        yield Context(executor=dask_executor)
        dask_executor.close()
    elif request.param == "delayed_dist":
        with distributed.Client(
                n_workers=2,
                threads_per_worker=4,
                processes=True
        ) as _:
            yield Context(executor=DelayedJobExecutor())
    elif request.param == "dask_make_default":
        try:
            ctx = Context.make_with(
                'dask-make-default',
                cpus=4,
            )
            yield ctx
        finally:
            # cleanup: Close cluster and client
            # This is also tested below, here just to make
            # sure things behave as expected.
            assert isinstance(ctx.executor, DaskJobExecutor)
            ctx.executor.is_local = True
            ctx.close()
    elif request.param == "dask_integration":
        with distributed.Client(
                n_workers=2,
                threads_per_worker=4,
                processes=False
        ) as _:
            yield Context.make_with("dask-integration")
    elif request.param == "concurrent":
        yield Context.make_with("threads")
    elif request.param == "delayed":
        yield Context(executor=DelayedJobExecutor())
    elif request.param == "pipelined":
        ctx = None
        try:
            ctx = Context.make_with('pipelined', cpus=range(2))
            yield ctx
        finally:
            if ctx is not None:
                ctx.close()


@pytest.fixture(
    scope='function',
    params=[
        # Only testing a selection of backends to keep the number of tests under control
        "HDF5", "RAW", "memory", SCIPY_COO, SPARSE_COO, CUPY, CUPY_SCIPY_CSR, "NPY", "BLO",
        "DM", "EMPAD", "FRMS6", "K2IS",
        "MIB", "MRC", "SEQ", "SER", "TVIPS",
        "RAW_CSR",
    ]
)
def load_kwargs(request, hdf5, default_raw, default_npy, default_npy_filepath, raw_csr_generated):
    param = request.param
    testdata_path = get_testdata_path()
    if param == 'HDF5':
        return {
            'filetype': 'HDF5',
            'path': hdf5.filename,
        }
    elif param == 'RAW':
        return {
            'filetype': 'RAW',
            'path': default_raw._path,
            'nav_shape': default_raw.shape.nav,
            'sig_shape': default_raw.shape.sig,
            'dtype': default_raw.dtype,
        }
    elif param == 'memory':
        return {
            'filetype': 'memory',
            'data': np.ones((3, 4, 5, 6)),
        }
    elif param in BACKENDS:
        if param in CUPY_BACKENDS and not has_cupy:
            pytest.skip("No CuPy, skipping CuPy-based tiles")
        return {
            'filetype': 'memory',
            'data': np.ones((3, 4, 5, 6)),
            'array_backends': (param, NUMPY)
        }
    elif param == 'NPY':
        return {
            'filetype': 'NPY',
            'path': default_npy_filepath,
        }
    elif param == 'BLO':
        blo_path = os.path.join(testdata_path, 'default.blo')
        if os.path.isfile(blo_path):
            return {
                'filetype': 'BLO',
                'path': blo_path,
            }
    elif param == 'DM':
        dm_files = list(sorted(glob(os.path.join(testdata_path, 'dm', '2018-7-17*.dm4'))))
        if dm_files:
            return {
                'filetype': 'dm',
                'files': dm_files
            }
    elif param == 'EMPAD':
        empad_path = os.path.join(testdata_path, 'EMPAD', 'acquisition_12_pretty.xml')
        if os.path.isfile(empad_path):
            return {
                'filetype': 'EMPAD',
                'path': empad_path
            }
    elif param == 'FRMS6':
        frms6_path = os.path.join(testdata_path, 'frms6', 'C16_15_24_151203_019.hdr')
        if os.path.isfile(frms6_path):
            return {
                'filetype': 'frms6',
                'path': frms6_path
            }
    elif param == 'K2IS':
        k2is_path = os.path.join(testdata_path, 'Capture52', 'Capture52_.gtg')
        if os.path.isfile(k2is_path):
            return {
                'filetype': 'k2is',
                'path': k2is_path
            }
    elif param == 'MIB':
        mib_path = os.path.join(testdata_path, 'default.mib')
        if os.path.isfile(mib_path):
            return {
                'filetype': 'mib',
                'path': mib_path,
                'nav_shape': (32, 32)
            }
    elif param == 'MRC':
        mrc_path = os.path.join(testdata_path, 'mrc', '20200821_92978_movie.mrc')
        if os.path.isfile(mrc_path):
            return {
                'filetype': 'mrc',
                'path': mrc_path
            }
    elif param == 'SEQ':
        seq_path = os.path.join(testdata_path, 'default.seq')
        if os.path.isfile(seq_path):
            return {
                'filetype': 'seq',
                'path': seq_path,
                'nav_shape': (8, 8)
            }
    elif param == 'SER':
        ser_path = os.path.join(testdata_path, 'default.ser')
        if os.path.isfile(ser_path):
            return {
                'filetype': 'ser',
                'path': ser_path
            }
    elif param == 'TVIPS':
        tvips_path = os.path.join(testdata_path, 'TVIPS', 'rec_20200623_080237_000.tvips')
        if os.path.isfile(tvips_path):
            return {
                'filetype': 'TVIPS',
                'path': tvips_path
            }
    elif param == 'RAW_CSR':
        if os.path.isfile(raw_csr_generated._path):
            return {
                'filetype': 'raw_csr',
                'path': raw_csr_generated._path,
            }
    else:
        raise ValueError(f'Unknown file type {param}')
    pytest.skip(f"Data file not found for {param}")


def _make_udfs(ds):
    def factory():
        m = np.zeros(ds.shape.sig)
        m[-1, -1] = 1.3
        return m

    udfs = [
        StdDevUDF(), ApplyMasksUDF(mask_factories=[factory])
    ]
    return udfs


def _calculate(ctx, load_kwargs):
    print(f"calculating with {ctx.executor}")
    ds = ctx.load(**load_kwargs)
    udfs = _make_udfs(ds)
    roi = np.zeros(
        np.prod(ds.shape.nav, dtype=np.int64),
        dtype=bool
    )
    roi[0] = True
    roi[-1] = True
    roi[len(roi)//2] = True
    roi = roi.reshape(ds.shape.nav)
    print(f"calculating {load_kwargs['filetype']}")
    result = ctx.run_udf(
        dataset=ds, udf=udfs, roi=roi, progress=True,
    )
    return result, udfs


@pytest.fixture(scope='function')
def reference(load_kwargs):
    # This must be function-scoped because load_kwargs is
    # Could refactor the tests to avoid this recalculate if needed
    ctx = Context(executor=InlineJobExecutor())
    return _calculate(ctx, load_kwargs)


@pytest.mark.slow
def test_executors(ctx, load_kwargs, reference):
    result, udfs = _calculate(ctx, load_kwargs)
    print(f"filetype: {load_kwargs['filetype']}")

    for i, item in enumerate(reference[0]):
        print(f"{i}: UDF {udfs[i]}")
        assert item.keys() == result[i].keys()
        for buf_key in item.keys():
            print(f"buffer {buf_key}")
            left = item[buf_key].raw_data
            right = np.array(result[i][buf_key].raw_data)
            print(np.max(np.abs(left - right)))
            print(np.min(np.abs(left)))
            print(np.min(np.abs(right)))
            # To see what type we actually test
            print("is allclose", np.allclose(left, right))
            print("dtypes", left.dtype, right.dtype)
            assert np.allclose(left, right), \
                f"mismatching result for buffer {buf_key} in UDF {udfs[i]}"


@pytest.mark.slow
def test_make_default():
    try:
        ctx = Context.make_with("dask-make-default")
        # This queries Dask which scheduler it is using
        ctx2 = Context.make_with("dask-integration")
        # make sure the second uses the Client of the first
        assert ctx2.executor.client is ctx.executor.client
    finally:
        # dask-make-default starts a Client that will persist
        # and not be closed automatically. We have to make sure
        # to close everything ourselves
        if ctx.executor.client.cluster is not None:
            ctx.executor.client.cluster.close(timeout=30)
        ctx.executor.client.close()
        ctx.close()


@pytest.mark.slow
def test_make_dask_implicit():
    with Context.make_with(cpus=(4, 7)) as ctx:
        assert ctx.executor.__class__ == DaskJobExecutor
        assert ctx.executor.client is not None
        assert len(ctx.executor.get_available_workers().has_cpu()) == 2


def test_connect_default(local_cluster_url):
    try:
        executor = DaskJobExecutor.connect(
            local_cluster_url,
            client_kwargs={'set_as_default': True}
        )
        ctx = Context(executor=executor)
        # This queries Dask which scheduler it is using
        ctx2 = Context.make_with("dask-integration")
        # make sure the second uses the Client of the first
        assert ctx2.executor.client is ctx.executor.client
    finally:
        # Only close the Client, keep the cluster running
        # since that is test infrastructure
        executor.client.close()
        ctx.close()


def test_use_distributed():
    # This Client is pretty cheap to start
    # since it only uses threads
    with distributed.Client(
        n_workers=1, threads_per_worker=1, processes=False
    ) as c:
        ctx = Context.make_with("dask-integration")
        assert isinstance(ctx.executor, DaskJobExecutor)
        assert ctx.executor.client is c


def test_no_dangling_client():
    # Within the whole test suite and LiberTEM we should not have
    # a dangling dask.distributed Client set as default Dask scheduler.
    # That means we confirm that we get a ConcurrentJobExecutor in the
    # default case.
    ctx = Context.make_with("dask-integration")
    assert isinstance(ctx.executor, ConcurrentJobExecutor)


def test_use_threads():
    with dask.config.set(scheduler="threads"):
        ctx = Context.make_with("dask-integration")
        assert isinstance(ctx.executor, ConcurrentJobExecutor)
        assert isinstance(
            ctx.executor.client,
            (
                concurrent.futures.ThreadPoolExecutor,
                multiprocessing.pool.ThreadPool,
            )
        )


def test_use_synchronous():
    with dask.config.set(scheduler="synchronous"):
        ctx = Context.make_with("dask-integration")
        assert isinstance(ctx.executor, InlineJobExecutor)


# not implemented in most executors, also only used for the clustered / caching
# dataset stuff, which we are not using
@pytest.mark.xfail
@pytest.mark.slow
def test_executor_run_each_partition(ctx: Context):
    ds = ctx.load("memory", data=np.random.randint(0, 1024, size=(16, 16, 128, 128)))
    partitions = ds.get_partitions()

    def _per_partition(p):
        assert p.__class__.__name__ == "MemPartition"
        return 42

    for res in ctx.executor.run_each_partition(partitions, _per_partition):
        assert res == 42


@pytest.mark.slow
def test_executor_map(ctx: Context):
    inp = [1, 2, 3]
    exp = [2, 3, 4]
    assert list(ctx.executor.map(lambda x: x + 1, inp)) == exp


@pytest.mark.slow
def test_executor_run_each_host(ctx: Context):
    res = ctx.executor.run_each_host(lambda x: 42 + x, 1)
    workers = ctx.executor.get_available_workers()
    assert set(res.keys()) == workers.hosts()
    for k, v in res.items():
        assert k in workers.hosts()
        assert v == 43


@pytest.mark.skipif(has_gpus, reason='Test to check error on no-GPU avail')
def test_make_with_no_gpu():
    with pytest.raises(ExecutorSpecException):
        Context.make_with('dask', gpus=2)


@pytest.mark.slow
def test_make_with_scenarios():
    """
    Comment from @uellue on #1443:

    I see several scenarios to use these parameters:

    1. CPU defaults are OK, but I want to hand-craft and optimize my GPU worker setup
    2. GPU defaults are OK/not applicable, but I need to optimize my CPU worker setup.
    3. Co-optimize CPU and GPU setup
    4. Disable all CPU or GPU workers.

    For 1. and 2. it would be good if the user doesn't have to specify the CPU or
    GPU setup to retain it at the defaults. For 4. it is easy to specify ...=0. For
    that reason it would be better, IMO, if leaving a parameter at "None" keeps it
    at the defaults.
    """
    # 1. CPU defaults are OK, but I want to hand-craft and optimize my GPU worker setup
    try:
        with Context.make_with('dask', gpus=2) as ctx:
            assert len(ctx.executor.get_available_workers().has_cpu()) > 0
            assert len(ctx.executor.get_available_workers().has_cuda()) == 2
    except ExecutorSpecException:
        assert not has_gpus, "should only happen if we have no GPUs available"

    # 2. GPU defaults are OK/not applicable, but I need to optimize my CPU worker setup.
    with Context.make_with('dask', cpus=2) as ctx:
        assert len(ctx.executor.get_available_workers().has_cpu()) == 2
        if has_gpus:
            assert len(ctx.executor.get_available_workers().has_cuda()) > 0

    # 3. Co-optimize CPU and GPU setup
    try:
        with Context.make_with('dask', cpus=2, gpus=2) as ctx:
            assert len(ctx.executor.get_available_workers().has_cpu()) > 0
            assert len(ctx.executor.get_available_workers().has_cuda()) == 2
    except ExecutorSpecException:
        assert not has_gpus, "should only happen if we have no GPUs available"

    # 4.1 Disable all CPU* or GPU workers.
    with Context.make_with('dask', cpus=0) as ctx:
        assert len(ctx.executor.get_available_workers().has_cpu()) == 0
        if has_gpus:
            assert len(ctx.executor.get_available_workers().has_cuda()) > 0
        else:
            # FIXME: this should raise, as we don't have any non-service
            # workers, correct?
            pass

    # 4.1 Disable all CPU or GPU* workers.
    # (this should even be possible if no GPUs are available anyways)
    with Context.make_with('dask', gpus=0) as ctx:
        assert len(ctx.executor.get_available_workers().has_cpu()) > 0
        assert len(ctx.executor.get_available_workers().has_cuda()) == 0


class MockPartition:
    @property
    def slice(self):
        return Slice(origin=(0, 0, 0), shape=Shape((16, 16, 16), sig_dims=1))


class MockTask:
    def __call__(self, params, env: Environment) -> Any:
        time.sleep(0.025 / 2)
        # print(f"MockTask.__call__: params={params}, env={env}")
        return params

    def get_tracing_span_context(self):
        raise NotImplementedError()

    def get_partition(self):
        return MockPartition()

    def get_locations(self):
        return None

    def get_resources(self):
        return {}


class DelayingCommHandler(TaskCommHandler):
    def handle_task(self, task: TaskProtocol, queue: WorkerQueue):
        # our tests only work if the tasks don't all get submitted at the
        # beginning of the `run_tasks` call - this simulates the live
        # processing scenario
        time.sleep(0.025)


@pytest.mark.parametrize('executor', ['dask', 'pipelined', 'inline', 'concurrent'])
def test_scatter_update(executor, local_cluster_ctx, pipelined_ctx, concurrent_executor):
    import uuid
    cancel_id = str(uuid.uuid4())

    # we do this dance to re-use the existing executors, so we don't have to
    # mark this test as slow:
    if executor == 'dask':
        ctx = local_cluster_ctx
    elif executor == 'pipelined':
        ctx = pipelined_ctx
    elif executor == 'inline':
        ctx = Context.make_with('inline')
    elif executor == 'concurrent':
        ctx = Context(executor=concurrent_executor)
    else:
        raise ValueError('invalid executor name')

    print(f"starting for {executor}: {ctx.executor} {cancel_id}")

    exc = ctx.executor
    comm_handler = DelayingCommHandler()

    # for concurrent, we need to take the number of threads into account,
    # otherwise the tasks will start running too soon for the first update
    # to take place
    num_workers = sum(w.nthreads for w in exc.get_available_workers())

    results = []

    with exc.scatter('hello scatter') as handle:
        result_iter = exc.run_tasks(
            cancel_id=cancel_id,
            params_handle=handle,
            task_comm_handler=comm_handler,
            tasks=[MockTask() for _ in range(30 * num_workers)]
        )
        first_result, _task = next(result_iter)
        print("started")
        assert first_result == "hello scatter"
        print(first_result)
        exc.scatter_update(handle, 'new value')
        for result, _task in result_iter:
            results.append(result)

    # eventually, the new value is available to the workers:
    assert "new value" in results

    print("done")


@pytest.mark.parametrize('executor', ['dask', 'pipelined', 'inline', 'concurrent'])
def test_scatter_patch(executor, local_cluster_ctx, pipelined_ctx, concurrent_executor):
    import uuid
    cancel_id = str(uuid.uuid4())

    # we do this dance to re-use the existing executors, so we don't have to
    # mark this test as slow:
    if executor == 'dask':
        ctx = local_cluster_ctx
    elif executor == 'pipelined':
        ctx = pipelined_ctx
    elif executor == 'inline':
        ctx = Context.make_with('inline')
    elif executor == 'concurrent':
        ctx = Context(executor=concurrent_executor)
    else:
        raise ValueError('invalid executor name')

    print(f"starting for {executor}: {ctx.executor} {cancel_id}")

    exc = ctx.executor
    comm_handler = DelayingCommHandler()

    # for concurrent, we need to take the number of threads into account,
    # otherwise the tasks will start running too soon for the first update
    # to take place
    num_workers = sum(w.nthreads for w in exc.get_available_workers())

    results = []

    class Patchable:
        def __init__(self, value):
            self.value = value

        def patch(self, patch):
            self.value = patch

        def __repr__(self):
            return f"<Patchable: {self.value}>"

    with exc.scatter(Patchable('hello scatter')) as handle:
        result_iter = exc.run_tasks(
            cancel_id=cancel_id,
            params_handle=handle,
            task_comm_handler=comm_handler,
            tasks=[MockTask() for _ in range(30 * num_workers)]
        )
        first_result, _task = next(result_iter)
        print("started")
        assert first_result.value == "hello scatter"
        print(first_result)
        exc.scatter_update_patch(handle, 'new value')
        for result, _task in result_iter:
            results.append(result.value)

    # eventually, the new value is available to the workers:
    assert "new value" in results

    print("done")
    print(results)



import os

import pytest
import numpy as np

from libertem.udf.sum import SumUDF
from libertem.io.dataset.ser import SERDataSet
from libertem.udf.sumsigudf import SumSigUDF
from libertem.io.dataset.base import TilingScheme
from libertem.common import Shape
from libertem.common.buffers import reshaped_view

from utils import dataset_correction_verification, get_testdata_path, ValidationUDF, roi_as_sparse

try:
    import hyperspy.api as hs
except ModuleNotFoundError:
    hs = None


SER_TESTDATA_PATH = os.path.join(get_testdata_path(), 'default.ser')
HAVE_SER_TESTDATA = os.path.exists(SER_TESTDATA_PATH)

pytestmark = pytest.mark.skipif(not HAVE_SER_TESTDATA, reason="need SER testdata")


@pytest.fixture
def default_ser(lt_ctx):
    ds = SERDataSet(path=SER_TESTDATA_PATH)
    ds.set_num_cores(4)
    ds = ds.initialize(lt_ctx.executor)
    assert tuple(ds.shape) == (8, 35, 512, 512)
    return ds


@pytest.fixture(scope='module')
def default_ser_raw():
    res = hs.load(str(SER_TESTDATA_PATH))
    return res.data


def test_smoke(lt_ctx):
    ds = lt_ctx.load("ser", path=SER_TESTDATA_PATH)
    p = next(ds.get_partitions())
    tileshape = Shape(
        (1,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )

    next(p.get_tiles(tiling_scheme))


@pytest.mark.skipif(hs is None, reason="No HyperSpy found")
def test_comparison(default_ser, default_ser_raw, lt_ctx_fast):
    udf = ValidationUDF(
        reference=reshaped_view(default_ser_raw, (-1, *tuple(default_ser.shape.sig)))
    )
    lt_ctx_fast.run_udf(udf=udf, dataset=default_ser)


@pytest.mark.skipif(hs is None, reason="No HyperSpy found")
def test_comparison_roi(default_ser, default_ser_raw, lt_ctx_fast):
    roi = np.random.choice(
        [True, False],
        size=tuple(default_ser.shape.nav),
        p=[0.5, 0.5]
    )
    udf = ValidationUDF(reference=default_ser_raw[roi])
    lt_ctx_fast.run_udf(udf=udf, dataset=default_ser, roi=roi)


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_roi(lt_ctx, as_sparse):
    ds = lt_ctx.load("ser", path=SER_TESTDATA_PATH)
    roi = np.zeros(ds.shape.nav, dtype=bool)
    roi[0, 1] = True

    ds.set_num_cores(2)
    parts = ds.get_partitions()

    p = next(parts)
    tileshape = Shape(
        (1,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )

    t0 = next(p.get_tiles(tiling_scheme, roi=roi))
    assert t0.tile_slice.origin[0] == 0

    p1 = next(parts)
    roi.reshape((-1,))[p1.slice.get(nav_only=True)] = True
    if as_sparse:
        roi = roi_as_sparse(roi)
    t1 = next(p1.get_tiles(tiling_scheme, roi=roi))
    assert t1.tile_slice.origin[0] == 1


def test_with_udf(lt_ctx):
    ds = lt_ctx.load("ser", path=SER_TESTDATA_PATH)
    udf = SumUDF()
    udf_results = lt_ctx.run_udf(udf=udf, dataset=ds)

    # compare result with a known-working variant:
    result = np.zeros(ds.shape.sig)
    for p in ds.get_partitions():
        tileshape = Shape(
            (1,) + tuple(ds.shape.sig),
            sig_dims=ds.shape.sig.dims
        )
        tiling_scheme = TilingScheme.make_for_shape(
            tileshape=tileshape,
            dataset_shape=ds.shape,
        )
        for tile in p.get_tiles(tiling_scheme):
            result[tile.tile_slice.get(sig_only=True)] += np.sum(tile.data, axis=0)

    assert np.allclose(udf_results['intensity'].data, result)


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
@pytest.mark.slow
def test_correction(lt_ctx, with_roi):
    ds = lt_ctx.load("ser", path=SER_TESTDATA_PATH)

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None

    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


def test_positive_sync_offset(default_ser, lt_ctx):
    udf = SumSigUDF()
    sync_offset = 2

    ds_with_offset = SERDataSet(
        path=SER_TESTDATA_PATH, sync_offset=sync_offset
    )
    ds_with_offset.set_num_cores(4)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == 2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (1,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    t0 = next(p0.get_tiles(tiling_scheme))
    assert tuple(t0.tile_slice.origin) == (0, 0, 0)

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (210, 0, 0)
    assert p.slice.shape[0] == 70

    result = lt_ctx.run_udf(dataset=default_ser, udf=udf)
    result = result['intensity'].raw_data[sync_offset:]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[
        :ds_with_offset._meta.image_count - sync_offset
    ]

    assert np.allclose(result, result_with_offset)


def test_negative_sync_offset(default_ser, lt_ctx):
    udf = SumSigUDF()
    sync_offset = -2

    ds_with_offset = SERDataSet(
        path=SER_TESTDATA_PATH, sync_offset=sync_offset
    )
    ds_with_offset.set_num_cores(4)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == -2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (1,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    t0 = next(p0.get_tiles(tiling_scheme))
    assert tuple(t0.tile_slice.origin) == (2, 0, 0)

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (210, 0, 0)
    assert p.slice.shape[0] == 70

    result = lt_ctx.run_udf(dataset=default_ser, udf=udf)
    result = result['intensity'].raw_data[:default_ser._meta.image_count - abs(sync_offset)]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[abs(sync_offset):]

    assert np.allclose(result, result_with_offset)


def test_positive_sync_offset_with_roi(default_ser, lt_ctx):
    udf = SumSigUDF()
    result = lt_ctx.run_udf(dataset=default_ser, udf=udf)
    result = result['intensity'].raw_data

    sync_offset = 2

    ds_with_offset = lt_ctx.load("ser", path=SER_TESTDATA_PATH, sync_offset=sync_offset)

    roi = np.random.choice([False], (8, 35))
    roi[0:1] = True

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf, roi=roi)
    result_with_offset = result_with_offset['intensity'].raw_data

    assert np.allclose(result[sync_offset:35 + sync_offset], result_with_offset)


def test_negative_sync_offset_with_roi(default_ser, lt_ctx):
    udf = SumSigUDF()
    result = lt_ctx.run_udf(dataset=default_ser, udf=udf)
    result = result['intensity'].raw_data

    sync_offset = -2

    ds_with_offset = lt_ctx.load(
        "ser", path=SER_TESTDATA_PATH, sync_offset=sync_offset
    )

    roi = np.random.choice([False], (8, 35))
    roi[0:1] = True

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf, roi=roi)
    result_with_offset = result_with_offset['intensity'].raw_data

    assert np.allclose(result[:35 + sync_offset], result_with_offset[abs(sync_offset):])


def test_missing_frames(lt_ctx):
    nav_shape = (10, 35)

    ds = SERDataSet(path=SER_TESTDATA_PATH, nav_shape=nav_shape)
    ds.set_num_cores(4)
    ds = ds.initialize(lt_ctx.executor)

    tileshape = Shape(
        (1,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )

    for p in ds.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p._start_frame == 262
    assert p._num_frames == 88
    assert p.slice.origin == (262, 0, 0)
    assert p.slice.shape[0] == 88
    assert t.tile_slice.origin == (279, 0, 0)
    assert t.tile_slice.shape[0] == 1


def test_too_many_frames(lt_ctx):
    nav_shape = (7, 35)

    ds = SERDataSet(path=SER_TESTDATA_PATH, nav_shape=nav_shape)
    ds.set_num_cores(4)
    ds = ds.initialize(lt_ctx.executor)

    tileshape = Shape(
        (1,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )

    for p in ds.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass


def test_offset_smaller_than_image_count(lt_ctx):
    sync_offset = -286

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "ser",
            path=SER_TESTDATA_PATH,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-280, 280\), which is \(-image_count, image_count\)"
    )


def test_offset_greater_than_image_count(lt_ctx):
    sync_offset = 286

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "ser",
            path=SER_TESTDATA_PATH,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-280, 280\), which is \(-image_count, image_count\)"
    )


def test_reshape_nav(default_ser, lt_ctx):
    udf = SumSigUDF()

    ds_with_1d_nav = lt_ctx.load("ser", path=SER_TESTDATA_PATH, nav_shape=(8,))
    result_with_1d_nav = lt_ctx.run_udf(dataset=ds_with_1d_nav, udf=udf)
    result_with_1d_nav = result_with_1d_nav['intensity'].raw_data

    ds_with_2d_nav = lt_ctx.load("ser", path=SER_TESTDATA_PATH, nav_shape=(4, 2))
    result_with_2d_nav = lt_ctx.run_udf(dataset=ds_with_2d_nav, udf=udf)
    result_with_2d_nav = result_with_2d_nav['intensity'].raw_data

    ds_with_3d_nav = lt_ctx.load("ser", path=SER_TESTDATA_PATH, nav_shape=(2, 2, 2))
    result_with_3d_nav = lt_ctx.run_udf(dataset=ds_with_3d_nav, udf=udf)
    result_with_3d_nav = result_with_3d_nav['intensity'].raw_data

    assert np.allclose(result_with_1d_nav, result_with_2d_nav, result_with_3d_nav)


def test_incorrect_sig_shape(lt_ctx):
    sig_shape = (5, 5)

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "ser",
            path=SER_TESTDATA_PATH,
            sig_shape=sig_shape
        )
    assert e.match(
        r"sig_shape must be of size: 262144"
    )


def test_scheme_too_large(default_ser):
    partitions = default_ser.get_partitions()
    p = next(partitions)
    depth = p.shape[0]

    # we make a tileshape that is too large for the partition here:
    tileshape = Shape(
        (depth + 1,) + tuple(default_ser.shape.sig),
        sig_dims=default_ser.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_ser.shape,
    )

    # tile shape is clamped to partition shape
    # (in SER adjusted to depth=1 at the moment):
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape)[0] <= depth


def test_bad_params(ds_params_tester, standard_bad_ds_params):
    args = ("ser", SER_TESTDATA_PATH)
    for params in standard_bad_ds_params:
        ds_params_tester(*args, **params)



import numpy as np
import inspect
from .base import BaseAnalysis, AnalysisResult, AnalysisResultSet
from libertem.analysis.getroi import get_roi
from libertem.udf.sum import SumUDF
from .helper import GeneratorHelper


class SumTemplate(GeneratorHelper):

    short_name = "sum"

    def __init__(self, params):
        self.params = params

    def get_dependency(self):
        return [
            "from matplotlib import colors",
            "from libertem import masks",
            "from libertem.udf.sum import SumUDF"
        ]

    def get_docs(self):
        title = "SUM Analysis"
        docs_rst = inspect.getdoc(SumUDF)
        docs = self.format_docs(title, docs_rst)
        return docs

    def get_analysis(self):
        roi = self.get_roi()
        temp_analysis = [
                f"{roi}",
                "sum_udf = SumUDF()",
                "sum_result = ctx.run_udf(dataset=ds, udf=sum_udf, roi=roi)"
                ]
        return '\n'.join(temp_analysis)

    def get_plot(self):
        plot = [
            "plt.figure()",
            "plt.imshow(sum_result['intensity'], norm=colors.LogNorm())",
            "plt.colorbar()",
         ]
        return ['\n'.join(plot)]


class SumResultSet(AnalysisResultSet):
    """
    Running a :class:`SumAnalysis` via :meth:`libertem.api.Context.run`
    returns an instance of this class.

    If the dataset contains complex numbers, the regular result attribute carries the
    absolute value of the result and additional attributes with real part, imaginary part,
    phase and full complex result are available.

    .. versionadded:: 0.3.0

    Attributes
    ----------
    intensity : libertem.analysis.base.AnalysisResult
        Sum of all detector frames along the navigation dimension,
        preserving the signal dimension. Absolute value of the sum if the dataset
        contains complex numbers. Log-scaled visualization.
    intensity_lin : libertem.analysis.base.AnalysisResult
        Sum of all detector frames along the navigation dimension,
        preserving the signal dimension. Absolute value of the sum if the dataset
        contains complex numbers. Lin-scaled visualization.

        .. versionadded:: 0.6.0

    intensity_real : libertem.analysis.base.AnalysisResult
        Real part of the sum of all detector frames along the navigation dimension,
        preserving the signal dimension. This is only available if the dataset
        contains complex numbers.
    intensity_imag : libertem.analysis.base.AnalysisResult
        Imaginary part of the sum of all detector frames along the navigation dimension,
        preserving the signal dimension. This is only available if the dataset
        contains complex numbers.
    intensity_angle : libertem.analysis.base.AnalysisResult
        Phase angle of the sum of all detector frames along the navigation dimension,
        preserving the signal dimension. This is only available if the dataset
        contains complex numbers.
    intensity_complex : libertem.analysis.base.AnalysisResult
        Complex result of the sum of all detector frames along the navigation dimension,
        preserving the signal dimension. This is only available if the dataset
        contains complex numbers.
    """
    pass


class SumAnalysis(BaseAnalysis, id_="SUM_FRAMES"):
    TYPE = 'UDF'

    def get_udf(self):
        dest_dtype = np.dtype(self.dataset.dtype)
        if dest_dtype.kind not in ('c', 'f'):
            dest_dtype = 'float32'
        return SumUDF(dtype=dest_dtype)

    def get_roi(self):
        return get_roi(params=self.parameters, shape=self.dataset.shape.nav)

    def get_udf_results(self, udf_results, roi, damage):
        from libertem.viz import visualize_simple
        if udf_results['intensity'].data.dtype.kind == 'c':
            return AnalysisResultSet(
                self.get_complex_results(
                    udf_results['intensity'].data,
                    key_prefix="intensity",
                    title="intensity",
                    desc="sum of all frames",
                    damage=True,
                    default_lin=False,
                )
            )

        return SumResultSet([
            AnalysisResult(
                raw_data=udf_results['intensity'].data,
                visualized=visualize_simple(
                    udf_results['intensity'].data, logarithmic=True,
                    damage=True
                ),
                key="intensity", title="intensity [log]", desc="sum of frames log-scaled"
            ),
            AnalysisResult(raw_data=udf_results['intensity'].data,
                visualized=visualize_simple(
                    udf_results['intensity'].data, logarithmic=False,
                    damage=True
                ),
                key="intensity_lin", title="intensity [lin]", desc="sum of frames lin-scaled"
            ),
        ])

    @classmethod
    def get_template_helper(cls):
        return SumTemplate



import os
import logging
from typing import Optional
import warnings
import contextlib

import numpy as np
from ncempy.io.ser import fileSER
from sparseconverter import CUDA, NUMPY, ArrayBackend

from libertem.common.math import prod, flat_nonzero
from libertem.common import Shape, Slice
from libertem.io.dataset.base.tiling_scheme import TilingScheme
from libertem.common.messageconverter import MessageConverter
from .base import (
    DataSet, FileSet, BasePartition, DataSetException, DataSetMeta,
    DataTile,
)

log = logging.getLogger(__name__)


class SERDatasetParams(MessageConverter):
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/SERDatasetParams.schema.json",
        "title": "SERDatasetParams",
        "type": "object",
        "properties": {
            "type": {"const": "SER"},
            "path": {"type": "string"},
            "nav_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sig_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sync_offset": {"type": "number"},
        },
        "required": ["type", "path"]
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path"]
        }
        if "nav_shape" in raw_data:
            data["nav_shape"] = tuple(raw_data["nav_shape"])
        if "sig_shape" in raw_data:
            data["sig_shape"] = tuple(raw_data["sig_shape"])
        if "sync_offset" in raw_data:
            data["sync_offset"] = raw_data["sync_offset"]
        return data


class SERFile:
    def __init__(self, path, num_frames):
        self._path = path
        self._num_frames = num_frames

    def _get_handle(self):
        return fileSER(self._path)

    @contextlib.contextmanager
    def get_handle(self):
        with self._get_handle() as f:
            yield f

    @property
    def num_frames(self):
        return self._num_frames

    @property
    def start_idx(self):
        return 0

    @property
    def end_idx(self):
        return self.num_frames


class SERFileSet(FileSet):
    pass


class SERDataSet(DataSet):
    """
    Read TIA SER files.

    Examples
    --------

    >>> ds = ctx.load("ser", path="/path/to/file.ser")  # doctest: +SKIP

    Parameters
    ----------
    path: str
        Path to the .ser file

    nav_shape: tuple of int, optional
        A n-tuple that specifies the size of the navigation region ((y, x), but
        can also be of length 1 for example for a line scan, or length 3 for
        a data cube, for example)

    sig_shape: tuple of int, optional
        Signal/detector size (height, width)

    sync_offset: int, optional
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start
    """
    def __init__(self, path, emipath=None, nav_shape=None,
                 sig_shape=None, sync_offset=0, io_backend=None):
        super().__init__(io_backend=io_backend)
        if io_backend is not None:
            raise ValueError("SERDataSet currently doesn't support alternative I/O backends")
        self._path = path
        self._meta = None
        self._filesize = None
        self._num_frames = None
        if emipath is not None:
            warnings.warn(
                "emipath is not used anymore, as it was removed from ncempy", DeprecationWarning
            )
        self._nav_shape = tuple(nav_shape) if nav_shape else nav_shape
        self._sig_shape = tuple(sig_shape) if sig_shape else sig_shape
        self._sync_offset = sync_offset

    def _do_initialize(self):
        self._filesize = os.stat(self._path).st_size
        reader = SERFile(path=self._path, num_frames=None)

        with reader.get_handle() as f1:
            self._num_frames = f1.head['ValidNumberElements']
            if f1.head['ValidNumberElements'] == 0:
                raise DataSetException("no data found in file")

            data, meta_data = f1.getDataset(0)
            dtype = f1._dictDataType[meta_data['DataType']]
            nav_dims = tuple(
                reversed([
                    int(dim['DimensionSize'])
                    for dim in f1.head['Dimensions']
                ])
            )
            self._image_count = int(self._num_frames)
            if self._nav_shape is None:
                self._nav_shape = nav_dims
            if self._sig_shape is None:
                self._sig_shape = tuple(data.shape)
            elif int(prod(self._sig_shape)) != int(prod(data.shape)):
                raise DataSetException(
                    "sig_shape must be of size: %s" % int(prod(data.shape))
                )
            self._nav_shape_product = int(prod(self._nav_shape))
            self._sync_offset_info = self.get_sync_offset_info()
            self._shape = Shape(self._nav_shape + self._sig_shape, sig_dims=len(self._sig_shape))
            self._meta = DataSetMeta(
                shape=self._shape,
                raw_dtype=dtype,
                sync_offset=self._sync_offset,
                image_count=self._image_count,
            )
        return self

    def initialize(self, executor):
        return executor.run_function(self._do_initialize)

    @classmethod
    def get_msg_converter(cls):
        return SERDatasetParams

    @classmethod
    def get_supported_extensions(cls):
        return {"ser"}

    @classmethod
    def get_supported_io_backends(self):
        return []

    @classmethod
    def detect_params(cls, path, executor):
        if path.lower().endswith(".ser"):
            ds = cls(path)
            ds = ds.initialize(executor)
            return {
                "parameters": {
                    "path": path,
                    "nav_shape": tuple(ds.shape.nav),
                    "sig_shape": tuple(ds.shape.sig),
                },
                "info": {
                    "image_count": int(prod(ds.shape.nav)),
                    "native_sig_shape": tuple(ds.shape.sig),
                }
            }
        return False

    @property
    def dtype(self):
        return self._meta.raw_dtype

    @property
    def shape(self):
        return self._meta.shape

    def check_valid(self):
        try:
            with fileSER(self._path) as f1:
                if f1.head['ValidNumberElements'] == 0:
                    raise DataSetException("no data found in file")
                if f1.head['DataTypeID'] not in (0x4120, 0x4122):
                    raise DataSetException("unknown datatype id: %s" % f1.head['DataTypeID'])
            return True
        except OSError as e:
            raise DataSetException("invalid dataset: %s" % e) from e

    def get_cache_key(self):
        return {
            "path": self._path,
            "shape": tuple(self.shape),
            "sync_offset": self._sync_offset,
        }

    def _get_fileset(self):
        assert self._num_frames is not None
        return SERFileSet([
            SERFile(
                path=self._path,
                num_frames=self._num_frames,
            )
        ])

    def get_base_shape(self, roi):
        return (1,) + tuple(self.shape.sig)

    def adjust_tileshape(self, tileshape, roi):
        # force single-frame tiles
        return (1,) + tileshape[1:]

    def get_partitions(self):
        fileset = self._get_fileset()
        for part_slice, start, stop in self.get_slices():
            yield SERPartition(
                path=self._path,
                meta=self._meta,
                partition_slice=part_slice,
                fileset=fileset,
                start_frame=start,
                num_frames=stop - start,
                io_backend=self.get_io_backend(),
                decoder=None,
            )

    def __repr__(self):
        return f"<SERDataSet for {self._path}>"


class SERPartition(BasePartition):
    def __init__(self, path, *args, **kwargs):
        self._path = path
        super().__init__(*args, **kwargs)

    def validate_tiling_scheme(self, tiling_scheme):
        if tiling_scheme.shape.sig != self.shape.sig:
            raise ValueError(
                f"invalid tiling scheme ({tiling_scheme.shape!r}): sig shape must match"
            )

    def _preprocess(self, tile_data, tile_slice):
        if self._corrections is None:
            return
        self._corrections.apply(tile_data, tile_slice)

    def get_tiles(self, tiling_scheme: TilingScheme, dest_dtype="float32", roi=None,
            array_backend: Optional[ArrayBackend] = None):
        if array_backend is None:
            array_backend = self.meta.array_backends[0]
        assert array_backend in (NUMPY, CUDA)
        sync_offset = self.meta.sync_offset
        shape = Shape((1,) + tuple(self.shape.sig), sig_dims=self.shape.sig.dims)
        tiling_scheme = tiling_scheme.adjust_for_partition(self)
        self.validate_tiling_scheme(tiling_scheme)

        start = self._start_frame
        if start < self.meta.image_count:
            stop = min(start + self._num_frames, self.meta.image_count)
            if roi is None:
                indices = np.arange(max(0, start), stop)
                # in case of a negative sync_offset, 'start' can be negative
                if start < 0:
                    offset = abs(sync_offset)
                else:
                    offset = start - sync_offset
            else:
                # The following is taken (effectively) from _default_get_read_ranges
                roi_nonzero = flat_nonzero(roi)

                shifted_roi = roi_nonzero + sync_offset
                roi_mask = np.logical_and(shifted_roi >= max(0, start),
                                          shifted_roi < stop)
                indices = shifted_roi[roi_mask]

                # in case of a negative sync_offset, 'start' can be negative
                if start < 0:
                    offset = np.sum(roi_nonzero < abs(sync_offset))
                else:
                    offset = np.sum(roi_nonzero < start - sync_offset)

            with fileSER(self._path) as f:
                for num, idx in enumerate(indices):
                    origin = (num + offset,) + tuple([0] * self.shape.sig.dims)
                    tile_slice = Slice(origin=origin, shape=shape)
                    data, metadata = f.getDataset(int(idx))
                    if data.dtype != np.dtype(dest_dtype):
                        data = data.astype(dest_dtype)
                    data = data.reshape(shape)
                    self._preprocess(data, tile_slice)
                    yield DataTile(
                        data,
                        tile_slice=tile_slice,
                        scheme_idx=0,
                    )



import numpy as np
import pytest

import libertem.udf.logsum as logsum
from sparseconverter import (
    SPARSE_BACKENDS, for_backend, NUMPY, SPARSE_COO, get_device_class
)

from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random, set_device_class


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(logsum.LogsumUDF().get_backends())
)
def test_logsum(lt_ctx, delayed_ctx, backend):
    with set_device_class(get_device_class(backend)):
        if backend in SPARSE_BACKENDS:
            data = _mk_random((16, 8, 32, 64), array_backend=SPARSE_COO)
        else:
            data = _mk_random((16, 8, 32, 64), array_backend=NUMPY)

        dataset = MemoryDataSet(
            data=data, tileshape=(8, 32, 64),
            num_partitions=2,
            sig_dims=2,
            array_backends=(backend, ) if backend is not None else None
        )

        logsum_result = logsum.run_logsum(ctx=lt_ctx, dataset=dataset)
        logsum_delayed = logsum.run_logsum(ctx=delayed_ctx, dataset=dataset)

        minima = np.min(data, axis=(2, 3))
        naive_result = for_backend(
            np.log(data - minima[..., np.newaxis, np.newaxis] + 1).sum(axis=(0, 1)),
            NUMPY
        )

        assert np.allclose(logsum_result['logsum'].data, naive_result)
        assert np.allclose(logsum_delayed['logsum'].data, naive_result)



import pathlib
import importlib
import numpy as np
import pytest
import dask.array as da

from libertem.udf.sumsigudf import SumSigUDF
from libertem.udf.sum import SumUDF
from libertem.io.dataset.base import DataSetException
from libertem.io.dataset.dask import DaskDataSet

from utils import dataset_correction_verification
# Load the dask dataset utils from this same folder
# This is really ugly but necessary to ensure that
# dask-worker processes can find utils_dask.py
location = pathlib.Path(__file__).parent / "utils_dask.py"
spec = importlib.util.spec_from_file_location("utils_dask", location)
utils_dask = importlib.util.module_from_spec(spec)
spec.loader.exec_module(utils_dask)
_mk_dask_from_delayed = utils_dask._mk_dask_from_delayed


@pytest.mark.parametrize(
    "dest_dtype", ('float32', 'int32')
)
def test_get_macrotile(lt_ctx, dest_dtype):
    data = _mk_dask_from_delayed(shape=(16, 16, 16, 16), chunking=(1, -1, -1, -1))
    ds = lt_ctx.load('dask', data, sig_dims=2)
    p = next(ds.get_partitions())
    mt = p.get_macrotile(dest_dtype=dest_dtype)
    assert tuple(mt.shape) == (256, 16, 16)
    assert mt.dtype == np.dtype(dest_dtype)


def test_merge_sig(lt_ctx):
    data = _mk_dask_from_delayed(shape=(5, 25, 16, 16), chunking=(1, -1, 8, 8))
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=0.)
    assert tuple(ds.array.chunks) == ((1,) * data.shape[0], (25,), (16,), (16,))


def test_contig_nav(lt_ctx):
    data = _mk_dask_from_delayed(shape=(5, 25, 16, 16), chunking=(2, 5, -1, -1))
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=0.)
    assert tuple(ds.array.chunks) == ((2, 2, 1), (25,), (16,), (16,))


def test_reorient_nav(lt_ctx):
    data = _mk_dask_from_delayed(shape=(5, 25, 16, 16), chunking=(5, 1, -1, -1))
    sig_dims = 2
    ds = lt_ctx.load('dask', data, sig_dims=sig_dims, preserve_dimensions=False, min_size=0.)
    assert tuple(ds.array.chunks) == ((1,) * data.shape[1], (5,), (16,), (16,))
    assert ds.array.shape == (25, 5, 16, 16)
    assert tuple(reversed(sorted(len(c) for c in ds.array.chunks[:-sig_dims])))\
           == tuple(len(c) for c in ds.array.chunks[:-sig_dims])


def test_reorient_nav2(lt_ctx):
    data = _mk_dask_from_delayed(shape=(5, 25, 16, 16), chunking=(1, -1, -1, -1))
    sig_dims = 2
    ds = lt_ctx.load('dask', data, sig_dims=sig_dims, preserve_dimensions=False, min_size=0.)
    assert tuple(ds.array.chunks) == ((1,) * data.shape[0], (25,), (16,), (16,))
    assert ds.array.shape == data.shape


def test_reorient_nav3(lt_ctx):
    data = _mk_dask_from_delayed(shape=(5, 25, 16, 16), chunking=(2, 1, -1, -1))
    sig_dims = 2
    ds = lt_ctx.load('dask', data, sig_dims=sig_dims, preserve_dimensions=True, min_size=0.)
    assert ds.array.shape == data.shape


@pytest.mark.dist
def test_size_based_merging(dist_ctx):
    dtype = np.float32
    data = _mk_dask_from_delayed(shape=(5, 25, 16, 16), chunking=(1, -1, -1, -1), dtype=dtype)
    chunksize = np.prod(data.shape[1:]) * np.dtype(dtype).itemsize
    min_size = chunksize * 2 - 1
    ds = dist_ctx.load('dask', data, sig_dims=2, min_size=min_size)
    assert tuple(ds.array.chunks) == ((3, 2), (25,), (16,), (16,))


def test_size_based_merging2(lt_ctx):
    dtype = np.float32
    data = _mk_dask_from_delayed(shape=(12, 25, 16, 16), chunking=(1, 2, -1, -1), dtype=dtype)
    frame_size = np.prod(data.shape[2:]) * np.dtype(dtype).itemsize
    min_size = frame_size * 4
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=min_size)
    assert tuple(ds.array.chunks) == ((1,) * data.shape[0], (6, 4, 4, 4, 7), (16,), (16,))


def test_size_based_merging3(lt_ctx):
    dtype = np.float32
    data = _mk_dask_from_delayed(shape=(12, 25, 16, 16), chunking=(1, 2, -1, -1), dtype=dtype)
    min_size = 0  # a zero minimum size should have no size-based merging
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=min_size)
    assert tuple(ds.array.chunks) == data.chunks


def test_size_based_merging4(lt_ctx):
    dtype = np.float32
    data = _mk_dask_from_delayed(shape=(6, 10, 16, 16), chunking=(1, -1, -1, -1), dtype=dtype)
    min_size = np.inf  # should result in one large partition
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=min_size)
    assert tuple(ds.array.chunks) == tuple((el,) for el in data.shape)


def test_no_chunking(lt_ctx):
    dtype = np.float32
    data = _mk_dask_from_delayed(shape=(5, 25, 16, 16), chunking=(-1, -1, -1, -1), dtype=dtype)
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=0.)
    assert tuple(ds.array.chunks) == ((5,), (25,), (16,), (16,))


def test_check_if_dask(lt_ctx):
    with pytest.raises(DataSetException):
        lt_ctx.load('dask', False, sig_dims=2)


def test_chunking_defined(lt_ctx):
    data = da.random.random(size=(5, 25, 16, 16), chunks=((-1,) * 4))
    data += 0.1
    mask = data.sum(axis=(1, 2, 3)) > 0.
    data = data[mask, ...]

    with pytest.raises(DataSetException):
        lt_ctx.load('dask', data, sig_dims=2)


def test_check_sig_dims(lt_ctx):
    data = da.random.random(size=(5, 25, 16, 16), chunks=((-1,) * 4))

    with pytest.raises(DataSetException):
        lt_ctx.load('dask', data, sig_dims=3.)


def test_check_sig_dims2(lt_ctx):
    data = da.random.random(size=(5, 25, 16, 16), chunks=((-1,) * 4))

    with pytest.raises(DataSetException):
        lt_ctx.load('dask', data, sig_dims=7)


def test_io_backend():
    data = da.random.random(size=(5, 25, 16, 16), chunks=((-1,) * 4))
    with pytest.raises(DataSetException):
        DaskDataSet(data, sig_dims=2, io_backend=True)


def test_get_num_part(lt_ctx):
    data = _mk_dask_from_delayed(shape=(5, 25, 16, 16), chunking=(2, 5, -1, -1))
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=0.)
    assert ds.get_num_partitions() == 3


def test_3d_array(lt_ctx):
    data = _mk_dask_from_delayed(shape=(5, 16, 16), chunking=(1, 8, 8))
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=0.)
    assert tuple(ds.array.chunks) == ((1,) * data.shape[0], (16,), (16,))


def test_4d_1sig_array(lt_ctx):
    data = _mk_dask_from_delayed(shape=(5, 36, 50, 16), chunking=(1, 6, 10, 8))
    ds = lt_ctx.load('dask', data, sig_dims=1, min_size=0.)
    assert tuple(ds.array.chunks) == ((1,) * 5, (6,) * 6, (50,), (16,))


def test_4d_3sig_array(lt_ctx):
    data = _mk_dask_from_delayed(shape=(5, 36, 50, 16), chunking=(1, 6, 10, 8))
    ds = lt_ctx.load('dask', data, sig_dims=3, min_size=0.)
    assert tuple(ds.array.chunks) == ((1,) * 5, (36,), (50,), (16,))


def test_5d_array(lt_ctx):
    data = _mk_dask_from_delayed(shape=(8, 4, 5, 16, 16), chunking=(4, 2, 1, 8, 8))
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=0.)
    assert tuple(ds.array.chunks) == ((4, 4), (4,), (5,), (16,), (16,))


def test_5d_size_based_merging(lt_ctx):
    dtype = np.float32
    data = _mk_dask_from_delayed(shape=(8, 12, 25, 16, 16), chunking=(1, 1, 2, -1, -1), dtype=dtype)
    frame_size = np.prod(data.shape[2:]) * np.dtype(dtype).itemsize
    min_size = frame_size * 4
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=min_size)
    assert tuple(ds.array.chunks) == ((1,) * data.shape[0], (4, 4, 4), (25,), (16,), (16,))


def test_6d_array(lt_ctx):
    data = _mk_dask_from_delayed(shape=(4, 8, 4, 5, 16, 16), chunking=(2, 4, 2, 1, 8, 8))
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=0.)
    assert tuple(ds.array.chunks) == ((2, 2), (8,), (4,), (5,), (16,), (16,))


def test_sum_udf(lt_ctx):
    data = _mk_dask_from_delayed(shape=(5, 25, 16, 16), chunking=(1, -1, 8, 8))
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=0.)
    res = lt_ctx.run_udf(ds, udf=SumUDF())
    assert np.allclose(res['intensity'].data, data.sum(axis=(0, 1)).compute())


def test_sumsig_udf(lt_ctx):
    data = _mk_dask_from_delayed(shape=(5, 25, 16, 16), chunking=(2, -1, 8, 8))
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=0.)
    res = lt_ctx.run_udf(ds, udf=SumSigUDF())
    assert np.allclose(res['intensity'].data, data.sum(axis=(2, 3)).compute())


def test_part_file_mapping(lt_ctx):
    data = _mk_dask_from_delayed(shape=(5, 25, 16, 16),
                                 chunking=(1, -1, -1, -1))
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=0.)

    for part_idx, part in enumerate(ds.get_partitions()):
        macrotile = part.get_macrotile(dest_dtype="float32")
        assert np.unique(np.asarray(macrotile.data)) == np.asarray([part_idx]).astype(np.float32)


def test_part_file_mapping2(lt_ctx):
    data = _mk_dask_from_delayed(shape=(4, 24, 16, 16),
                                 chunking=(2, 12, -1, -1))
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=0.)

    sequence = np.asarray([0, 1]).astype(np.float32)
    for part in ds.get_partitions():
        macrotile = part.get_macrotile(dest_dtype="float32")
        assert np.allclose(np.unique(np.asarray(macrotile.data)), sequence)
        sequence += sequence.size


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
def test_correction(lt_ctx, with_roi):
    data = _mk_dask_from_delayed(shape=(5, 25, 16, 16), chunking=(2, -1, -1, -1))
    ds = lt_ctx.load('dask', data, sig_dims=2, min_size=0.)

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None

    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


@pytest.mark.dist
def test_dist_process(dist_ctx):
    shape = (3, 5, 7, 11)
    data = da.random.random(shape, chunks=(2, 2, 2, 2))
    roi = np.random.choice([True, False], data.shape[:2])
    # Dask doesn't do multi-dimensional fancy indexing with booleans,
    # unlike NumPy :rolleyes:
    ref = data.reshape((np.prod(shape[:2]), *shape[2:]))[roi.flatten()].sum(axis=0).compute()
    ds = dist_ctx.load("dask", data, sig_dims=2)
    res = dist_ctx.run_udf(dataset=ds, udf=SumUDF(), roi=roi)
    assert np.allclose(res['intensity'].raw_data, ref)



import numpy as np
import pytest

from libertem.common.math import prod, make_2D_square
from libertem.common.shape import Shape


@pytest.mark.parametrize(
    ('sequence', 'ref', 'typ'), [
        ([], 1, int),
        (np.array([]), 1, int),
        ((1, 2, 3), 6, int),
        ((-11, 2, 3), -66, int),
        ((1., 2, False), 0, ValueError),
        (np.array((1., 1+2j, 1), dtype=np.complex128), 1, ValueError),
        ((2**32, 2**32, 2**32), 2**96, int),
        (np.array((2**62, 2**62, 2**62), dtype=np.int64), 2**(3*62), int),
        (Shape((1, 2, 3), sig_dims=1).nav, 2, int)
    ]
)
def test_prod(sequence, ref, typ):
    if typ is int:
        res = prod(sequence)
        assert res == ref
        assert isinstance(res, typ)
    else:
        with pytest.raises(typ):
            res = prod(sequence)


@pytest.mark.parametrize(
    ('shape', 'result'), [
        ((16,), (4, 4)),
        ((100,), (10, 10)),
        ((15,), (15,)),
        ((1,), (1, 1)),
        ((-20,), ValueError),
        ((3, 7), (3, 7)),
        (tuple(), tuple()),
        ((0.4,), ValueError),
    ]
)
def test_make_2D_square(shape, result):
    if isinstance(result, tuple):
        assert make_2D_square(shape) == result
    elif issubclass(result, Exception):
        with pytest.raises(result):
            make_2D_square(shape)
            return



import logging
from typing import TYPE_CHECKING

import tornado.web
from opentelemetry import trace

from libertem.web.engine import JobEngine
from .base import CORSMixin, log_message
from .state import SharedState
from .messages import Message


if TYPE_CHECKING:
    from .events import EventRegistry

log = logging.getLogger(__name__)
tracer = trace.get_tracer(__name__)


class JobDetailHandler(CORSMixin, tornado.web.RequestHandler):
    def initialize(self, state: SharedState, event_registry: "EventRegistry"):
        self.state = state
        self.event_registry = event_registry
        self.engine = JobEngine(state, event_registry)

    async def put(self, job_id):
        with tracer.start_as_current_span("JobDetailHandler.put") as span:
            request_data = tornado.escape.json_decode(self.request.body)
            analysis_id = request_data['job']['analysis']
            await self.engine.register_job(analysis_id, job_id)
            serialized_job = self.state.job_state.serialize(job_id)
            msg = Message().start_job(
                serialized_job=serialized_job, analysis_id=analysis_id,
            )
            span.set_attributes({
                "libertem.job_id": job_id,
                "libertem.analysis_id": analysis_id,
            })
            log_message(msg)
            self.write(msg)
            self.finish()
            await self.engine.run_analysis(analysis_id, job_id)

    async def delete(self, job_id):
        result = await self.state.job_state.remove(job_id)
        if result:
            msg = Message().cancel_job(job_id)
            log_message(msg)
            self.event_registry.broadcast_event(msg)
            self.write(msg)
        else:
            log.warning("tried to remove unknown job %s", job_id)
            msg = Message().cancel_failed(job_id)
            log_message(msg)
            self.event_registry.broadcast_event(msg)
            self.write(msg)



import h5py
import numpy as np
from PIL import Image

from .base import ResultFormat


class HDF5ResultFormat(ResultFormat):
    @classmethod
    def format_info(cls):
        return {
            "id": "HDF5",
            "description": "HDF5 container (.h5)",
        }

    def serialize_to_buffer(self, buf):
        with h5py.File(buf, 'w') as f:
            for k in self.get_result_keys():
                f[k] = self._result_set[k]
            # FIXME: add "metadata", for example, what analysis was run that
            # resulted in this file

    def get_content_type(self):
        return "application/x-hdf5"

    def get_filename(self):
        return "results.h5"  # FIXME: more specific name


class NPZResultFormat(ResultFormat):
    @classmethod
    def format_info(cls):
        return {
            "id": "NPZ",
            "description": "numpy format (.npz)",
        }

    def _get_result_dict(self):
        return {
            k: np.array(self._result_set[k])
            for k in self.get_result_keys()
        }

    def serialize_to_buffer(self, buf):
        np.savez(buf, **self._get_result_dict())

    def get_content_type(self):
        return "application/octet-stream"

    def get_filename(self):
        return "results.npz"  # FIXME: naming


class NPZCompressedResultFormat(NPZResultFormat):
    @classmethod
    def format_info(cls):
        return {
            "id": "NPZ_COMPRESSED",
            "description": "numpy format, compressed (.npz)",
        }

    def serialize_to_buffer(self, buf):
        np.savez_compressed(buf, **self._get_result_dict())


class TiffResultFormat(ResultFormat):
    @classmethod
    def format_info(cls):
        return {
            "id": "TIFF",
            "description": "Multi-page 32bit float TIFF (.tif)",
        }

    def get_channel_images(self):
        for k in self.get_result_keys():
            result = np.array(self._result_set[k]).astype(np.float32)
            yield Image.fromarray(result)

    def serialize_to_buffer(self, buf):
        images = self.get_channel_images()
        first_image = next(images)
        first_image.save(buf, format="TIFF", save_all=True, append_images=images)

    def get_content_type(self):
        return "image/tiff"

    def get_filename(self):
        return "results.tif"  # FIXME: naming


class RawResultFormat(ResultFormat):
    @classmethod
    def format_info(cls):
        return {
            "id": "RAW",
            "description": "Raw binary, as-is (.bin)",
        }

    def _get_result_arr(self):
        return np.stack([
            np.array(self._result_set[k])
            for k in self.get_result_keys()
        ])

    def serialize_to_buffer(self, buf):
        buf.write(self._get_result_arr().tobytes())

    def get_content_type(self):
        return "application/octet-stream"

    def get_filename(self):
        arr = self._get_result_arr()
        return "results_{}_{}.bin".format(arr.dtype, "-".join(str(i) for i in arr.shape))



import numpy as np

from libertem.io.dataset.memory import MemoryDataSet
from libertem.io.dataset.base import TilingScheme
from libertem.common import Shape

from utils import _mk_random


def test_partition3d_correct_slices():
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(3, 16, 16),
                            num_partitions=2, sig_dims=2)

    tileshape = Shape(
        (3,) + tuple(dataset.shape.sig),
        sig_dims=dataset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=dataset.shape,
    )

    mask = np.zeros(data.shape[:2], dtype=bool)
    mask[0, 0] = True
    mask[15, 0] = True

    partitions = dataset.get_partitions()
    p1 = next(partitions)
    p2 = next(partitions)

    assert len(list(p1.get_tiles(tiling_scheme=tiling_scheme, roi=mask))) == 1
    assert len(list(p2.get_tiles(tiling_scheme=tiling_scheme, roi=mask))) == 1

    t1 = next(p1.get_tiles(tiling_scheme=tiling_scheme, roi=mask))
    t2 = next(p2.get_tiles(tiling_scheme=tiling_scheme, roi=mask))

    print("t1", t1.tile_slice)
    print("t2", t2.tile_slice)

    assert t1.tile_slice.origin[0] == 0
    assert t2.tile_slice.origin[0] == 1



# Was moved to common for MIT license compatibility
# in dataset code

from libertem.common.exceptions import (
    UDFException, ExecutorSpecException, UDFRunCancelled
)

__all__ = ["UDFException", "ExecutorSpecException", "UDFRunCancelled"]
